from typing import Literal

from pydantic import Field
from pydantic.json_schema import SkipJsonSchema
from rich.console import Group
from rich.text import Text
from typing_extensions import override

from pipelex.builder.pipe.pipe_spec import PipeSpec
from pipelex.pipe_operators.func.pipe_func_blueprint import PipeFuncBlueprint
from pipelex.tools.misc.pretty import PrettyPrintable


class PipeFuncSpec(PipeSpec):
    """PipeFunc enables calling custom functions in the Pipelex framework."""

    type: SkipJsonSchema[Literal["PipeFunc"]] = "PipeFunc"
    pipe_category: SkipJsonSchema[Literal["PipeOperator"]] = "PipeOperator"
    function_name: str = Field(description="The name of the function to call.")

    @override
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        # Get base pipe information from parent
        base_group = super().rendered_pretty(title=title, depth=depth)

        # Create a group combining base info with func-specific details
        func_group = Group()
        func_group.renderables.append(base_group)

        # Add function specific information
        func_group.renderables.append(Text())  # Blank line
        func_group.renderables.append(Text.from_markup(f"Function Name: [bold cyan]{self.function_name}[/bold cyan]"))

        return func_group

    @override
    def to_blueprint(self) -> PipeFuncBlueprint:
        base_blueprint = super().to_blueprint()
        return PipeFuncBlueprint(
            description=base_blueprint.description,
            inputs=base_blueprint.inputs,
            output=base_blueprint.output,
            function_name=self.function_name,
        )



from pipelex.cogt.exceptions import LLMConfigError
from pipelex.cogt.img_gen.img_gen_job_components import ImgGenJobConfig, ImgGenJobParams, ImgGenJobParamsDefaults
from pipelex.cogt.llm.llm_job_components import LLMJobConfig
from pipelex.cogt.models.model_deck_config import ModelDeckConfig
from pipelex.plugins.fal.fal_config import FalConfig
from pipelex.system.configuration.config_model import ConfigModel


class ExtractConfig(ConfigModel):
    page_output_text_file_name: str
    default_page_views_dpi: int


class ImgGenConfig(ConfigModel):
    img_gen_job_config: ImgGenJobConfig
    img_gen_param_defaults: ImgGenJobParamsDefaults
    fal_config: FalConfig

    def make_default_img_gen_job_params(self) -> ImgGenJobParams:
        return self.img_gen_param_defaults.make_img_gen_job_params()


class InstructorConfig(ConfigModel):
    is_dump_kwargs_enabled: bool
    is_dump_response_enabled: bool
    is_dump_error_enabled: bool


class LLMConfig(ConfigModel):
    instructor_config: InstructorConfig
    llm_job_config: LLMJobConfig
    is_structure_prompt_enabled: bool
    default_max_images: int
    is_dump_text_prompts_enabled: bool
    is_dump_response_text_enabled: bool
    generic_templates: dict[str, str]

    def get_template(self, template_name: str) -> str:
        template = self.generic_templates.get(template_name)
        if not template:
            msg = f"Template '{template_name}' not found in generic_templates"
            raise LLMConfigError(msg)
        return template


class Cogt(ConfigModel):
    model_deck_config: ModelDeckConfig
    llm_config: LLMConfig
    img_gen_config: ImgGenConfig
    extract_config: ExtractConfig



from typing import Literal

from pydantic import field_validator, model_validator
from typing_extensions import override

from pipelex.core.concepts.validation import is_concept_string_or_code_valid
from pipelex.core.pipes.pipe_blueprint import PipeBlueprint
from pipelex.pipe_controllers.sub_pipe_blueprint import SubPipeBlueprint
from pipelex.types import Self


class PipeParallelBlueprint(PipeBlueprint):
    type: Literal["PipeParallel"] = "PipeParallel"
    pipe_category: Literal["PipeController"] = "PipeController"
    parallels: list[SubPipeBlueprint]
    add_each_output: bool = False
    combined_output: str | None = None

    @property
    @override
    def pipe_dependencies(self) -> set[str]:
        """Return the set of pipe codes from the parallel branches."""
        return {parallel.pipe for parallel in self.parallels}

    @field_validator("combined_output", mode="before")
    @classmethod
    def validate_combined_output(cls, combined_output: str) -> str:
        if combined_output:
            if not is_concept_string_or_code_valid(concept_string_or_code=combined_output):
                msg = f"Combined output '{combined_output}' is not a valid concept string or code"
                raise ValueError(msg)
        return combined_output

    @model_validator(mode="after")
    def validate_output_options(self) -> Self:
        if not self.add_each_output and not self.combined_output:
            msg = (
                "PipeParallel requires either add_each_output to be True or combined_output to be set, "
                "or both, otherwise the pipe won't output anything"
            )
            raise ValueError(msg)
        return self

    @override
    def validate_inputs(self):
        pass

    @override
    def validate_output(self):
        pass



import sys
from typing import cast

from pydantic import Field, field_validator
from rich.console import Console
from rich.highlighter import Highlighter, JSONHighlighter, ReprHighlighter
from rich.logging import RichHandler

from pipelex.system.configuration.config_model import ConfigModel
from pipelex.system.console_target import ConsoleTarget
from pipelex.tools.log.log_levels import LogLevel
from pipelex.types import StrEnum


class LogMode(StrEnum):
    RICH = "rich"
    POOR = "poor"


class HighlighterName(StrEnum):
    JSON = "json"
    REPR = "repr"


class ProblemIds(StrEnum):
    AZURE_OPENAI_NO_STREAM_OPTIONS = "Azure OpenAI no stream_options"


class CallerInfoTemplate(StrEnum):
    FILE_LINE = "file_line"
    FILE_LINE_FUNC = "file_line_func"
    FUNC = "func"
    FILE_FUNC = "file_func"
    FUNC_LINE = "func_line"
    FUNC_MODULE = "func_module"
    FUNC_MODULE_LINE = "func_module_line"

    @classmethod
    def for_template_key(cls, key: "CallerInfoTemplate") -> str:
        match key:
            case cls.FILE_LINE:
                return "{file}:{line}"
            case cls.FILE_LINE_FUNC:
                return "{file}:{line} {func}"
            case cls.FUNC:
                return "{func}"
            case cls.FILE_FUNC:
                return "{file} {func}"
            case cls.FUNC_LINE:
                return "{func} {line}"
            case cls.FUNC_MODULE:
                return "{func} {module}"
            case cls.FUNC_MODULE_LINE:
                return "{func} {module} {line}"


class RichLogConfig(ConfigModel):
    is_show_time: bool
    is_show_level: bool
    is_link_path_enabled: bool
    highlighter_name: HighlighterName = Field(strict=False)
    is_markup_enabled: bool
    is_rich_tracebacks: bool
    is_tracebacks_word_wrap: bool
    is_tracebacks_show_locals: bool
    tracebacks_suppress: list[str]
    keywords_to_hilight: list[str]

    def make_rich_handler(self, target: ConsoleTarget) -> RichHandler:
        match target:
            case ConsoleTarget.STDOUT:
                console = Console(file=sys.stdout)
            case ConsoleTarget.STDERR:
                console = Console(file=sys.stderr)
            # case ConsoleTarget.FILE:
            #     console = Console(file=target.file_path)
            case _:
                msg = f"Invalid console target: {target}"
                raise ValueError(msg)
        highlighter: Highlighter
        match self.highlighter_name:
            case HighlighterName.JSON:
                highlighter = JSONHighlighter()
            case HighlighterName.REPR:
                highlighter = ReprHighlighter()

        return RichHandler(
            console=console,
            show_time=self.is_show_time,
            show_level=self.is_show_level,
            enable_link_path=self.is_link_path_enabled,
            highlighter=highlighter,
            markup=self.is_markup_enabled,
            rich_tracebacks=self.is_rich_tracebacks,
            tracebacks_word_wrap=self.is_tracebacks_word_wrap,
            tracebacks_show_locals=self.is_tracebacks_show_locals,
            tracebacks_suppress=self.tracebacks_suppress,
            keywords=self.keywords_to_hilight,
        )


class LogConfig(ConfigModel):
    default_log_level: LogLevel = Field(strict=False)
    package_log_levels: dict[str, LogLevel]
    log_mode: LogMode = Field(strict=False)
    console_log_target: ConsoleTarget = Field(strict=False)
    console_print_target: ConsoleTarget = Field(strict=False)

    is_console_logging_enabled: bool

    json_logs_indent: int
    presentation_line_width: int
    is_caller_info_enabled: bool
    caller_info_template: CallerInfoTemplate = Field(strict=False)

    silenced_problem_ids: list[str]

    rich_log_config: RichLogConfig

    # logger name to use for safe logging without fancy features like code filepath and stuff
    generic_poor_logger: str
    poor_loggers: list[str]

    @field_validator("package_log_levels", mode="before")
    @classmethod
    def validate_package_log_levels(cls, value: dict[str, str]) -> dict[str, LogLevel]:
        return cast(
            "dict[str, LogLevel]",
            ConfigModel.transform_dict_str_to_enum(input_dict=value, value_enum_cls=LogLevel),
        )



from pipelex.base_exceptions import PipelexError
from pipelex.core.memory.working_memory import WorkingMemory
from pipelex.types import Self


class PipeBuilderError(PipelexError):
    def __init__(self: Self, message: str, working_memory: WorkingMemory | None = None) -> None:
        self.working_memory = working_memory
        super().__init__(message)


class ConceptSpecError(PipelexError):
    """Details of a single concept failure during dry run."""


class PipelexBundleUnexpectedError(PipelexError):
    """Raised when an unexpected error occurs during validation."""



from pydantic import BaseModel
from typing_extensions import override

from pipelex import log
from pipelex.cogt.exceptions import LLMPromptParameterError
from pipelex.cogt.image.prompt_image import PromptImage
from pipelex.system.runtime import ProblemReaction, runtime_manager
from pipelex.tools.misc.string_utils import is_none_or_has_text, is_not_none_and_has_text


class LLMPrompt(BaseModel):
    system_text: str | None = None
    user_text: str | None = None
    user_images: list[PromptImage] = []

    def validate_before_execution(self):
        reaction = runtime_manager.problem_reactions.job
        match reaction:
            case ProblemReaction.NONE:
                pass
            case ProblemReaction.RAISE:
                if not is_none_or_has_text(text=self.system_text):
                    if self.system_text == "":
                        log.warning(f"system_text should be None or contain text. system_text = '{self.system_text}'")
                    else:
                        msg = "system_text should be None or contain text"
                        raise LLMPromptParameterError(msg)
                if not is_not_none_and_has_text(text=self.user_text):
                    msg = "user_text should contain text"
                    raise LLMPromptParameterError(msg)
            case ProblemReaction.LOG:
                if not is_none_or_has_text(text=self.system_text):
                    if self.system_text == "":
                        log.warning(f"system_text should be None or contain text. system_text = '{self.system_text}'")
                    else:
                        log.error(f"Prompt template system_text should be None or contain text. system_text = '{self.system_text}'")
                if not is_not_none_and_has_text(text=self.user_text):
                    log.error("user_text should contain text")

    @override
    def __str__(self) -> str:
        # return json_str(self, title="llm_prompt", is_spaced=True)
        return self.desc()
        # return "test"

    @override
    def __repr__(self) -> str:
        return self.desc()

    @override
    def __format__(self, format_spec: str) -> str:
        return self.desc()

    def desc(self, truncate_text_length: int | None = None) -> str:
        description = "LLM Prompt:"
        if truncate_text_length:
            if self.system_text:
                description += f"""
    system_text:
    {self.system_text[:truncate_text_length]}
    """
            if self.user_text:
                description += f"""
    user_text:
    {self.user_text[:truncate_text_length]}
    """
        else:
            if self.system_text:
                description += f"""
    system_text:
    {self.system_text}
    """
            if self.user_text:
                description += f"""
    user_text:
    {self.user_text}
    """
        if self.user_images:
            user_images_desc: str = "\n".join([f"  {image}" for image in self.user_images])

            description += f"""
user_images:
{user_images_desc}
"""
        return description



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import Annotated

import click
import typer
from posthog import tag
from rich.traceback import Traceback

from pipelex import log
from pipelex.cli.error_handlers import (
    ErrorContext,
    handle_model_availability_error,
    handle_model_choice_error,
    handle_model_deck_preset_error,
    handle_validate_bundle_error,
)
from pipelex.cogt.exceptions import ModelDeckPresetValidatonError
from pipelex.core.pipes.exceptions import PipeOperatorModelChoiceError
from pipelex.hub import get_console, get_library_manager, get_required_pipe, get_telemetry_manager, set_current_library
from pipelex.pipe_operators.exceptions import PipeOperatorModelAvailabilityError
from pipelex.pipe_run.dry_run import dry_run_pipe, dry_run_pipes
from pipelex.pipelex import Pipelex
from pipelex.pipeline.validate_bundle import ValidateBundleError, validate_bundle
from pipelex.system.runtime import IntegrationMode
from pipelex.system.telemetry.events import EventName, EventProperty
from pipelex.tools.misc.package_utils import get_package_version

COMMAND = "validate"


def do_validate_all_libraries_and_dry_run() -> None:
    try:
        with get_telemetry_manager().telemetry_context():
            tag(name=EventProperty.INTEGRATION, value=IntegrationMode.CLI)
            tag(name=EventProperty.PIPELEX_VERSION, value=get_package_version())
            tag(name=EventProperty.CLI_COMMAND, value=f"{COMMAND} all")

            library_manager = get_library_manager()
            library_id, library = library_manager.open_library()
            set_current_library(library_id=library_id)
            library_manager.load_libraries(library_id=library_id, library_dirs=[Path.cwd()])

            pipes = library.get_pipe_library().get_pipes()

            get_telemetry_manager().track_event(EventName.PIPE_DRY_RUN, properties={EventProperty.NB_PIPES: len(pipes)})

            asyncio.run(dry_run_pipes(pipes=pipes, raise_on_failure=True))
            log.info("Setup sequence passed OK, config and pipelines are validated.")
    except PipeOperatorModelAvailabilityError as exc:
        handle_model_availability_error(exc, context=ErrorContext.VALIDATION)
    except PipeOperatorModelChoiceError as exc:
        handle_model_choice_error(exc, context=ErrorContext.VALIDATION)


def validate_cmd(
    target: Annotated[
        str | None,
        typer.Argument(help="Pipe code, bundle file path (auto-detected based on .plx extension), or 'all' to validate all pipes"),
    ] = None,
    pipe: Annotated[
        str | None,
        typer.Option("--pipe", help="Pipe code to validate (optional when using --bundle)"),
    ] = None,
    bundle: Annotated[
        str | None,
        typer.Option("--bundle", help="Bundle file path (.plx) - validates all pipes in the bundle"),
    ] = None,
) -> None:
    """Validate and dry run a pipe or a bundle or all pipes.

    Examples:
        pipelex validate my_pipe
        pipelex validate my_bundle.plx
        pipelex validate --bundle my_bundle.plx
        pipelex validate --bundle my_bundle.plx --pipe my_pipe
        pipelex validate all
    """
    pipelex_instance: Pipelex
    # Check for "all" keyword
    if target == "all" and not pipe and not bundle:
        pipelex_instance = Pipelex.make(integration_mode=IntegrationMode.CLI)
        do_validate_all_libraries_and_dry_run()
        pipelex_instance.teardown()
        return

    # Validate mutual exclusivity
    provided_options = sum([target is not None, pipe is not None, bundle is not None])
    if provided_options == 0:
        ctx: click.Context = click.get_current_context()
        typer.echo(ctx.get_help())
        raise typer.Exit(0)

    # Let's analyze the options and determine what pipe code to use and if we need to load a bundle
    pipe_code: str | None = None
    bundle_path: str | None = None

    # Determine source:
    if target:
        if target.endswith(".plx"):
            bundle_path = target
            if bundle:
                typer.secho(
                    "Failed to validate: cannot use option --bundle if you're already passing a bundle file (.plx) as positional argument",
                    fg=typer.colors.RED,
                    err=True,
                )
                raise typer.Exit(1)
        else:
            pipe_code = target
            if pipe:
                typer.secho(
                    "Failed to validate: cannot use option --pipe if you're already passing a pipe code as positional argument",
                    fg=typer.colors.RED,
                    err=True,
                )
                raise typer.Exit(1)

    if bundle:
        assert not bundle_path, "bundle_path should be None at this stage if --bundle is provided"
        bundle_path = bundle

    if pipe:
        assert not pipe_code, "pipe_code should be None at this stage if --pipe is provided"
        pipe_code = pipe

    if not pipe_code and not bundle_path:
        typer.secho("Failed to validate: no pipe code or bundle file specified", fg=typer.colors.RED, err=True)
        raise typer.Exit(1)

    async def validate_pipe(pipe_code: str | None = None, bundle_path: str | None = None):
        if bundle_path:
            try:
                await validate_bundle(plx_file_path=bundle_path)
                typer.secho(f"✅ Successfully validated bundle '{bundle_path}'", fg=typer.colors.GREEN)
            except FileNotFoundError as exc:
                get_console().print(Traceback())
                typer.secho(f"Failed to load bundle '{bundle_path}':", fg=typer.colors.RED, err=True)
                raise typer.Exit(1) from exc
            except ValidateBundleError as bundle_error:
                handle_validate_bundle_error(bundle_error, bundle_path=bundle_path)
        elif pipe_code:
            typer.echo(f"Validating pipe '{pipe_code}'...")
            library_manager = get_library_manager()
            library_id, _ = library_manager.open_library()
            set_current_library(library_id=library_id)
            library_manager.load_libraries(library_id=library_id, library_dirs=[Path.cwd()])

            pipe = get_required_pipe(pipe_code=pipe_code)
            get_telemetry_manager().track_event(EventName.PIPE_DRY_RUN, properties={EventProperty.PIPE_TYPE: pipe.type})
            await dry_run_pipe(
                pipe,
                raise_on_failure=True,
            )
            typer.secho(f"✅ Successfully validated pipe '{pipe_code}'", fg=typer.colors.GREEN)
        else:
            typer.secho("Failed to validate: no pipe code or bundle specified", fg=typer.colors.RED, err=True)
            raise typer.Exit(1)

    try:
        pipelex_instance = Pipelex.make(integration_mode=IntegrationMode.CLI)
    except ModelDeckPresetValidatonError as model_deck_error:
        handle_model_deck_preset_error(model_deck_error, context=ErrorContext.VALIDATION)

    try:
        with get_telemetry_manager().telemetry_context():
            tag(name=EventProperty.INTEGRATION, value=IntegrationMode.CLI)
            tag(name=EventProperty.PIPELEX_VERSION, value=get_package_version())
            if bundle_path:
                tag(name=EventProperty.CLI_COMMAND, value=f"{COMMAND} bundle")
            else:
                tag(name=EventProperty.CLI_COMMAND, value=f"{COMMAND} pipe")

            asyncio.run(validate_pipe(pipe_code=pipe_code, bundle_path=bundle_path))
    except PipeOperatorModelChoiceError as exc:
        handle_model_choice_error(exc, context=ErrorContext.VALIDATION)
    except PipeOperatorModelAvailabilityError as exc:
        handle_model_availability_error(exc, context=ErrorContext.VALIDATION)
    finally:
        pipelex_instance.teardown()



"""UI components for routing profile configuration in the init command."""

from rich.console import Console, Group
from rich.panel import Panel
from rich.prompt import Prompt
from rich.table import Table
from rich.text import Text

from pipelex.tools.misc.string_utils import snake_to_capitalize_first_letter


def build_primary_backend_panel(backend_keys: list[str], backend_options: list[tuple[str, str]]) -> Panel:
    """Build a panel for selecting a primary backend from enabled backends.

    Args:
        backend_keys: List of enabled backend keys.
        backend_options: List of all backend options to get display names.

    Returns:
        A Panel containing the primary backend selection interface.
    """
    # Create a mapping of backend_key to display_name
    backend_key_to_name = dict(backend_options)

    table = Table(show_header=False, box=None, padding=(0, 2))
    table.add_column(style="bold cyan", justify="right", width=4)
    table.add_column(style="bold", width=30)

    for idx, backend_key in enumerate(backend_keys, start=1):
        display_name = backend_key_to_name.get(backend_key, snake_to_capitalize_first_letter(backend_key))
        table.add_row(f"[{idx}]", display_name)

    description = Text(
        "Multiple backends are enabled. Select which backend should be your primary/default.\n"
        "The primary backend will be used for models that don't have an exact match or pattern match in your routes.",
        style="dim",
    )

    return Panel(
        Group(description, Text(""), table),
        title="[bold yellow]Primary Backend Selection[/bold yellow]",
        border_style="yellow",
        padding=(1, 2),
    )


def prompt_primary_backend(console: Console, backend_keys: list[str]) -> str:
    """Prompt user to select a primary backend from enabled backends.

    Args:
        console: Rich Console instance for user interaction.
        backend_keys: List of enabled backend keys.

    Returns:
        The selected backend key.
    """
    # Default to first backend
    default_str = "1"

    selected_backend: str | None = None
    while selected_backend is None:
        choice_str = Prompt.ask("[bold]Enter your choice[/bold]", default=default_str, console=console)
        choice_input = choice_str.strip()

        # Parse input - handle empty (use default)
        if not choice_input:
            selected_backend = backend_keys[0]
            break

        try:
            # Parse as 1-based index from user input
            user_index = int(choice_input)

            # Validate index is in range (1-based)
            if user_index < 1 or user_index > len(backend_keys):
                max_idx = len(backend_keys)
                console.print(f"[red]Invalid choice: {user_index}.[/red] Please enter a number between 1 and {max_idx}.\n")
                continue

            # Convert to 0-based index and get backend key
            selected_backend = backend_keys[user_index - 1]
            break

        except ValueError:
            console.print(f"[red]Invalid input: '{choice_str}'.[/red] Please enter a number.\n")

    return selected_backend


def build_fallback_order_panel(remaining_backends: list[str], backend_options: list[tuple[str, str]]) -> Panel:
    """Build a panel explaining fallback order configuration.

    Args:
        remaining_backends: List of backend keys (excluding primary).
        backend_options: List of all backend options to get display names.

    Returns:
        A Panel containing the fallback order explanation and current order.
    """
    # Create a mapping of backend_key to display_name
    backend_key_to_name = dict(backend_options)

    table = Table(show_header=False, box=None, padding=(0, 2))
    table.add_column(style="bold cyan", justify="right", width=4)
    table.add_column(style="bold", width=30)

    for idx, backend_key in enumerate(remaining_backends, start=1):
        display_name = backend_key_to_name.get(backend_key, snake_to_capitalize_first_letter(backend_key))
        table.add_row(f"[{idx}]", display_name)

    description = Text(
        "Fallback order determines which backends to try when a model has no exact match or pattern match\n"
        "and is not available in your default backend.\nCurrent order:",
        style="dim",
    )

    return Panel(
        Group(description, Text(""), table),
        title="[bold yellow]Fallback Backend Order[/bold yellow]",
        border_style="yellow",
        padding=(1, 2),
    )


def prompt_fallback_order(console: Console, remaining_backends: list[str], backend_options: list[tuple[str, str]]) -> list[str]:
    """Prompt user to order the remaining backends for fallback.

    Args:
        console: Rich Console instance for user interaction.
        remaining_backends: List of backend keys (excluding primary).
        backend_options: List of all backend options to get display names.

    Returns:
        Ordered list of backend keys for fallback.
    """
    backend_key_to_name = dict(backend_options)

    console.print()
    console.print("[dim]Enter the order of backends (space or comma separated indices).[/dim]")
    console.print("[dim]For example: '2 1 3' or '2,1,3' to reorder them.[/dim]")
    console.print()

    fallback_order: list[str] | None = None
    while fallback_order is None:
        # Show current order as default
        default_order = " ".join(str(i + 1) for i in range(len(remaining_backends)))
        choice_str = Prompt.ask("[bold]Enter order[/bold]", default=default_order, console=console)
        choice_input = choice_str.strip()

        # Handle empty input (use default order)
        if not choice_input:
            fallback_order = remaining_backends.copy()
            break

        # Split by comma or space
        parts = choice_input.replace(",", " ").split()

        try:
            # Parse as 1-based indices from user input
            user_indices = [int(part.strip()) for part in parts if part.strip()]

            # Validate: must have exactly the same number of indices
            if len(user_indices) != len(remaining_backends):
                console.print(f"[red]Error: You must specify all {len(remaining_backends)} backends.[/red] You provided {len(user_indices)}.\n")
                continue

            # Validate: all indices are in range (1-based)
            invalid_indices = [i for i in user_indices if i < 1 or i > len(remaining_backends)]
            if invalid_indices:
                max_idx = len(remaining_backends)
                console.print(f"[red]Invalid choice(s): {invalid_indices}.[/red] Please enter numbers between 1 and {max_idx}.\n")
                continue

            # Validate: no duplicates
            if len(user_indices) != len(set(user_indices)):
                console.print("[red]Error: Duplicate indices not allowed.[/red] Each backend must appear exactly once.\n")
                continue

            # Convert to 0-based indices and reorder backends
            zero_based_indices = [i - 1 for i in user_indices]
            fallback_order = [remaining_backends[i] for i in zero_based_indices]
            break

        except ValueError:
            console.print(f"[red]Invalid input: '{choice_str}'.[/red] Please enter numbers separated by commas or spaces.\n")

    # Display the final order
    console.print("\n[green]✓[/green] Fallback order set to:")
    for idx, backend_key in enumerate(fallback_order, start=1):
        display_name = backend_key_to_name.get(backend_key, snake_to_capitalize_first_letter(backend_key))
        console.print(f"   {idx}. {display_name}")

    return fallback_order


def display_routing_profile_result(console: Console, profile_name: str, created: bool = False) -> None:
    """Display confirmation of routing profile configuration.

    Args:
        console: Rich Console instance for output.
        profile_name: Name of the routing profile that was set.
        created: Whether the profile was newly created.
    """
    if created:
        console.print(f"\n[green]✓[/green] Created and set routing profile to: [bold cyan]{profile_name}[/bold cyan]")
    else:
        console.print(f"\n[green]✓[/green] Routing profile set to: [bold cyan]{profile_name}[/bold cyan]")



from pipelex.cogt.usage.cost_category import CostCategory, CostsByCategoryDict


def model_cost_per_token(costs: CostsByCategoryDict, cost_category: CostCategory) -> float:
    # cost_per_million_tokens_usd should be missing only for models that we run on our own GPUs
    # all token types are not used for all models
    match cost_category:
        case CostCategory.INPUT_CACHED:
            if cost_per_million_tokens := costs.get(CostCategory.INPUT_CACHED):
                return cost_per_million_tokens / 1000000
            elif cost_per_million_tokens := costs.get(CostCategory.INPUT):
                # according to openai docs, cached input tokens are discounted 50%
                return 0.5 * cost_per_million_tokens / 1000000
            else:
                return 0.0
        case CostCategory.INPUT_NON_CACHED:
            return model_cost_per_token(costs=costs, cost_category=CostCategory.INPUT)
        case (
            CostCategory.INPUT
            | CostCategory.INPUT_JOINED
            | CostCategory.INPUT_AUDIO
            | CostCategory.OUTPUT
            | CostCategory.OUTPUT_AUDIO
            | CostCategory.OUTPUT_REASONING
            | CostCategory.OUTPUT_ACCEPTED_PREDICTION
            | CostCategory.OUTPUT_REJECTED_PREDICTION
        ):
            if cost_per_million_tokens := costs.get(cost_category):
                return cost_per_million_tokens / 1000000
            else:
                return 0.0



"""Topological sorting utilities for pipe dependencies."""

from pipelex.core.bundles.pipelex_bundle_blueprint import PipeBlueprintUnion
from pipelex.core.pipes.exceptions import PipeValidationError, PipeValidationErrorType


def sort_pipes_by_dependencies(
    pipes: dict[str, PipeBlueprintUnion],
) -> list[tuple[str, PipeBlueprintUnion]]:
    """Sort pipes by their dependencies using depth-first pre-order traversal.

    This performs a depth-first traversal where controllers appear before their dependencies,
    and dependencies are visited in the order specified by the controller (e.g., step order
    for PipeSequence). This ordering is intuitive for understanding pipeline structure.

    Args:
        pipes: Dictionary mapping pipe_code to PipeBlueprintUnion

    Returns:
        List of (pipe_code, pipe_blueprint) tuples sorted with controllers before dependencies.
        For PipeSequence, dependencies follow step order. For others, alphabetical order.

    Raises:
        PipeValidationError: If circular dependencies are detected among pipes

    Example:
        >>> pipes = {
        ...     "pipe_c": pipe_c_blueprint,  # sequence with steps: pipe_a, pipe_b
        ...     "pipe_a": pipe_a_blueprint,  # no dependencies
        ...     "pipe_b": pipe_b_blueprint,  # no dependencies
        ... }
        >>> sorted_pipes = sort_pipes_by_dependencies(pipes)
        >>> [code for code, _ in sorted_pipes]
        ['pipe_c', 'pipe_a', 'pipe_b']
    """
    # Find root pipes (those not depended upon by anyone)
    all_dependencies: set[str] = set()
    for pipe_blueprint in pipes.values():
        all_dependencies.update(pipe_blueprint.pipe_dependencies)

    root_pipes = [code for code in pipes if code not in all_dependencies]

    # Depth-first traversal to order pipes
    visited: set[str] = set()
    visiting: set[str] = set()  # For cycle detection
    sorted_pipes: list[tuple[str, PipeBlueprintUnion]] = []

    def visit(pipe_code: str) -> None:
        if pipe_code in visited:
            return
        if pipe_code in visiting:
            msg = f"Circular dependency detected involving pipe: {pipe_code}"
            raise PipeValidationError(message=msg, pipe_code=pipe_code, error_type=PipeValidationErrorType.CIRCULAR_DEPENDENCY_ERROR)
        if pipe_code not in pipes:
            # Dependency not in this bundle, skip it
            return

        visiting.add(pipe_code)
        pipe_blueprint = pipes[pipe_code]

        # Add current pipe first (pre-order)
        sorted_pipes.append((pipe_code, pipe_blueprint))

        # Visit dependencies in order
        ordered_deps = pipe_blueprint.ordered_pipe_dependencies
        if ordered_deps:
            # Use ordered dependencies (e.g., PipeSequence steps)
            for dep_code in ordered_deps:
                if dep_code in pipes:
                    visit(dep_code)
        else:
            # Use alphabetical order for determinism
            for dep_code in sorted(pipe_blueprint.pipe_dependencies):
                if dep_code in pipes:
                    visit(dep_code)

        visited.add(pipe_code)
        visiting.remove(pipe_code)

    # Start traversal from root pipes (sorted for determinism)
    for pipe_code in sorted(root_pipes):
        visit(pipe_code)

    # Ensure all pipes are included (handle disconnected components)
    for pipe_code in sorted(pipes.keys()):
        if pipe_code not in visited:
            visit(pipe_code)

    return sorted_pipes



from abc import ABC, abstractmethod

from pipelex.core.concepts.concept import Concept
from pipelex.core.concepts.native.concept_native import NativeConceptCode


class ConceptLibraryAbstract(ABC):
    @abstractmethod
    def add_new_concept(self, concept: Concept) -> None:
        pass

    @abstractmethod
    def add_concepts(self, concepts: list[Concept]) -> None:
        pass

    @abstractmethod
    def remove_concepts_by_concept_strings(self, concept_strings: list[str]) -> None:
        pass

    @abstractmethod
    def list_concepts_by_domain(self, domain: str) -> list[Concept]:
        pass

    @abstractmethod
    def list_concepts(self) -> list[Concept]:
        pass

    @abstractmethod
    def get_required_concept(self, concept_string: str) -> Concept:
        pass

    @abstractmethod
    def is_compatible(self, tested_concept: Concept, wanted_concept: Concept, strict: bool = False) -> bool:
        pass

    @abstractmethod
    def setup(self) -> None:
        pass

    @abstractmethod
    def reset(self) -> None:
        pass

    @abstractmethod
    def teardown(self) -> None:
        pass

    @abstractmethod
    def get_native_concept(self, native_concept: NativeConceptCode) -> Concept:
        pass

    @abstractmethod
    def get_required_concept_from_concept_string_or_code(self, concept_string_or_code: str, search_domains: list[str] | None = None) -> Concept:
        pass



from pipelex.cogt.img_gen.img_gen_job import ImgGenJob
from pipelex.cogt.img_gen.img_gen_job_components import ImgGenJobConfig, ImgGenJobParams, ImgGenJobReport
from pipelex.cogt.img_gen.img_gen_prompt import ImgGenPrompt
from pipelex.config import get_config
from pipelex.pipeline.job_metadata import JobCategory, JobMetadata


class ImgGenJobFactory:
    @classmethod
    def make_img_gen_job_from_prompt(
        cls,
        img_gen_prompt: ImgGenPrompt,
        img_gen_job_params: ImgGenJobParams | None = None,
        img_gen_job_config: ImgGenJobConfig | None = None,
        job_metadata: JobMetadata | None = None,
    ) -> ImgGenJob:
        img_gen_config = get_config().cogt.img_gen_config
        job_metadata = job_metadata or JobMetadata(
            job_category=JobCategory.IMG_GEN_JOB,
        )
        job_params = img_gen_job_params or img_gen_config.make_default_img_gen_job_params()
        job_config = img_gen_job_config or img_gen_config.img_gen_job_config
        job_report = ImgGenJobReport()

        return ImgGenJob(
            job_metadata=job_metadata,
            img_gen_prompt=img_gen_prompt,
            job_params=job_params,
            job_config=job_config,
            job_report=job_report,
        )

    @classmethod
    def make_img_gen_job_from_prompt_contents(
        cls,
        positive_text: str,
        img_gen_job_params: ImgGenJobParams | None = None,
        img_gen_job_config: ImgGenJobConfig | None = None,
        job_metadata: JobMetadata | None = None,
    ) -> ImgGenJob:
        img_gen_config = get_config().cogt.img_gen_config
        job_metadata = job_metadata or JobMetadata(
            job_category=JobCategory.IMG_GEN_JOB,
        )
        img_gen_prompt = ImgGenPrompt(positive_text=positive_text)
        job_params = img_gen_job_params or img_gen_config.make_default_img_gen_job_params()
        job_config = img_gen_job_config or img_gen_config.img_gen_job_config
        job_report = ImgGenJobReport()

        return ImgGenJob(
            job_metadata=job_metadata,
            img_gen_prompt=img_gen_prompt,
            job_params=job_params,
            job_config=job_config,
            job_report=job_report,
        )



from pydantic import BaseModel

from pipelex.types import StrEnum


class NodeCategory(StrEnum):
    SPECIAL = "special"
    STUFF = "stuff"
    CONDITION = "condition"


class NodeAttributeKey(StrEnum):
    CATEGORY = "category"
    NAME = "name"
    TAG = "tag"
    DESCRIPTION = "description"
    DEBUG_INFO = "debug_info"
    SUBGRAPH = "subgraph"
    COMMENT = "comment"


class SpecialNodeName(StrEnum):
    START = "start"


class EdgeCategory(StrEnum):
    PIPE = "pipe"
    BATCH = "batch"
    AGGREGATE = "aggregate"
    CONDITION = "condition"
    CHOICE = "choice"


class EdgeAttributeKey(StrEnum):
    EDGE_CATEGORY = "edge_category"
    PIPE_CODE = "pipe_code"
    CONDITION_EXPRESSION = "condition_expression"
    CHOSEN_PIPE = "chosen_pipe"


class GraphTree(BaseModel):
    nodes_by_subgraph: dict[str, list[str]]


class SubGraphClassDef(BaseModel):
    name: str
    color: str



from __future__ import annotations

from typing import Any

from pydantic import ConfigDict, Field, RootModel, model_validator
from typing_extensions import TYPE_CHECKING, override

from pipelex import log
from pipelex.cogt.exceptions import LLMPromptTemplateInputsError
from pipelex.system.runtime import ProblemReaction, runtime_manager
from pipelex.tools.misc.json_utils import json_str
from pipelex.tools.misc.string_utils import can_inject_text

if TYPE_CHECKING:
    from collections.abc import ItemsView

    from pipelex.types import Self

LLMPromptTemplateInputsDict = dict[str, Any]


class LLMPromptTemplateInputs(RootModel[LLMPromptTemplateInputsDict]):
    root: LLMPromptTemplateInputsDict = Field(default_factory=dict)

    model_config = ConfigDict(validate_assignment=True)

    @model_validator(mode="after")
    def validate_template_inputs(self) -> Self:
        if not self.root:
            # It's important to exit this func before calling get_config() because empty template inputs
            # are created during imports (before Pipelex is fully initialized)
            return self
        template_inputs_reaction = runtime_manager.problem_reactions.template_inputs
        if template_inputs_reaction == ProblemReaction.NONE:
            return self
        for var_name, value in self.root.items():
            if not can_inject_text(value):
                error_msg = f"Template input '{var_name}' cannot inject text"
                if template_inputs_reaction == ProblemReaction.RAISE:
                    raise LLMPromptTemplateInputsError(message=error_msg)
                log.error(error_msg)
        return self

    def items(self) -> ItemsView[str, Any]:
        return self.root.items()

    def list_keys(self) -> list[str]:
        return list(self.root.keys())

    def complemented_by(self, additional_template_inputs: LLMPromptTemplateInputs | None) -> LLMPromptTemplateInputs:
        all_template_inputs = self.root.copy()
        if additional_template_inputs:
            all_template_inputs.update(additional_template_inputs.root)
        return LLMPromptTemplateInputs(root=all_template_inputs)

    def complemented_by_dict(self, additional_inputs_dict: LLMPromptTemplateInputsDict | None) -> LLMPromptTemplateInputs:
        all_template_inputs = self.root.copy()
        if additional_inputs_dict:
            all_template_inputs.update(additional_inputs_dict)
        return LLMPromptTemplateInputs(root=all_template_inputs)

    @override
    def __str__(self) -> str:
        return json_str(self.root, title="llm_prompt_template_inputs", is_spaced=True)

    @staticmethod
    def from_args(**template_inputs: Any) -> LLMPromptTemplateInputs:
        return LLMPromptTemplateInputs(root=template_inputs)



from pipelex.base_exceptions import PipelexError
from pipelex.pipe_run.pipe_run_mode import PipeRunMode


class PipeRunParamsError(PipelexError):
    pass


class BatchParamsError(PipelexError):
    pass


class PipeRunError(PipelexError):
    def __init__(self, message: str, run_mode: PipeRunMode):
        self.run_mode = run_mode
        super().__init__(message)


class PipeRouterError(PipelexError):
    def __init__(
        self,
        message: str,
        run_mode: PipeRunMode,
        pipe_code: str,
        output_name: str | None,
        pipe_stack: list[str],
        missing_inputs: list[str] | None = None,
    ):
        self.run_mode = run_mode
        self.pipe_code = pipe_code
        self.output_name = output_name
        self.pipe_stack = pipe_stack
        self.missing_inputs = missing_inputs
        super().__init__(message)



from typing import Any

import openai
import shortuuid
from typing_extensions import override

from pipelex import log
from pipelex.cogt.exceptions import ImgGenGenerationError, SdkTypeError
from pipelex.cogt.image.generated_image import GeneratedImage
from pipelex.cogt.img_gen.img_gen_job import ImgGenJob
from pipelex.cogt.img_gen.img_gen_job_components import Quality
from pipelex.cogt.img_gen.img_gen_worker_abstract import ImgGenWorkerAbstract
from pipelex.cogt.model_backends.model_spec import InferenceModelSpec
from pipelex.plugins.openai.openai_img_gen_factory import OpenAIImgGenFactory
from pipelex.reporting.reporting_protocol import ReportingProtocol
from pipelex.tools.misc.base_64_utils import save_base_64_str_to_binary_file
from pipelex.tools.misc.file_utils import ensure_path

TEMP_OUTPUTS_DIR = "temp/img_gen_by_gpt_image"


class OpenAIImgGenWorker(ImgGenWorkerAbstract):
    def __init__(
        self,
        sdk_instance: Any,
        inference_model: InferenceModelSpec,
        reporting_delegate: ReportingProtocol | None = None,
    ):
        super().__init__(inference_model=inference_model, reporting_delegate=reporting_delegate)

        if not isinstance(sdk_instance, openai.AsyncOpenAI):
            msg = f"Provided ImgGen sdk_instance is not of type openai.AsyncOpenAI: it's a '{type(sdk_instance)}'"
            raise SdkTypeError(msg)

        self.openai_client = sdk_instance

    @override
    async def _gen_image(
        self,
        img_gen_job: ImgGenJob,
    ) -> GeneratedImage:
        one_image_list = await self._gen_image_list(img_gen_job=img_gen_job, nb_images=1)
        return one_image_list[0]

    @override
    async def _gen_image_list(
        self,
        img_gen_job: ImgGenJob,
        nb_images: int,
    ) -> list[GeneratedImage]:
        image_size = OpenAIImgGenFactory.image_size_for_gpt_image_1(aspect_ratio=img_gen_job.job_params.aspect_ratio)
        output_format = OpenAIImgGenFactory.output_format_for_gpt_image_1(output_format=img_gen_job.job_params.output_format)
        moderation = OpenAIImgGenFactory.moderation_for_gpt_image_1(is_moderated=img_gen_job.job_params.is_moderated)
        background = OpenAIImgGenFactory.background_for_gpt_image_1(background=img_gen_job.job_params.background)
        quality = OpenAIImgGenFactory.quality_for_gpt_image_1(quality=img_gen_job.job_params.quality or Quality.LOW)
        output_compression = OpenAIImgGenFactory.output_compression_for_gpt_image_1()
        result = await self.openai_client.images.generate(
            prompt=img_gen_job.img_gen_prompt.positive_text,
            model=self.inference_model.model_id,
            moderation=moderation,
            background=background,
            quality=quality,
            size=image_size,
            output_format=output_format,
            output_compression=output_compression,
            n=nb_images,
        )
        if not result.data:
            msg = "No result from OpenAI"
            raise ImgGenGenerationError(msg)

        generated_image_list: list[GeneratedImage] = []
        image_id = shortuuid.uuid()[:4]
        for image_index, image_data in enumerate(result.data):
            image_base64 = image_data.b64_json
            if not image_base64:
                msg = "No base64 image data received from OpenAI"
                raise ImgGenGenerationError(msg)

            folder_path = TEMP_OUTPUTS_DIR
            ensure_path(folder_path)
            img_path = f"{folder_path}/{image_id}_{image_index}.png"
            save_base_64_str_to_binary_file(base_64_str=image_base64, file_path=img_path)
            log.verbose(f"Saved image to {img_path}")
            generated_image_list.append(
                GeneratedImage(
                    url=img_path,
                    width=1024,
                    height=1024,
                ),
            )
        return generated_image_list



from pipelex import log
from pipelex.cogt.content_generation.assignment_models import ImgGenAssignment
from pipelex.cogt.image.generated_image import GeneratedImage
from pipelex.cogt.img_gen.img_gen_job_factory import ImgGenJobFactory
from pipelex.hub import get_img_gen_worker


async def img_gen_single_image(img_gen_assignment: ImgGenAssignment) -> GeneratedImage:
    img_gen_worker = get_img_gen_worker(img_gen_handle=img_gen_assignment.img_gen_handle)
    img_gen_job = ImgGenJobFactory.make_img_gen_job_from_prompt(
        img_gen_prompt=img_gen_assignment.img_gen_prompt,
        img_gen_job_params=img_gen_assignment.img_gen_job_params,
        img_gen_job_config=img_gen_assignment.img_gen_job_config,
        job_metadata=img_gen_assignment.job_metadata,
    )
    generated_image = await img_gen_worker.gen_image(img_gen_job=img_gen_job)
    log.verbose(f"generated_image:\n{generated_image}")
    return generated_image


async def img_gen_image_list(img_gen_assignment: ImgGenAssignment) -> list[GeneratedImage]:
    img_gen_worker = get_img_gen_worker(img_gen_handle=img_gen_assignment.img_gen_handle)
    img_gen_job = ImgGenJobFactory.make_img_gen_job_from_prompt(
        img_gen_prompt=img_gen_assignment.img_gen_prompt,
        img_gen_job_params=img_gen_assignment.img_gen_job_params,
        img_gen_job_config=img_gen_assignment.img_gen_job_config,
        job_metadata=img_gen_assignment.job_metadata,
    )
    generated_image_list = await img_gen_worker.gen_image_list(
        img_gen_job=img_gen_job,
        nb_images=img_gen_assignment.nb_images,
    )
    log.verbose(f"generated_image_list:\n{generated_image_list}")
    return generated_image_list


async def img_gen_image(img_gen_assignment: ImgGenAssignment) -> GeneratedImage | list[GeneratedImage]:
    if img_gen_assignment.nb_images > 1:
        return await img_gen_image_list(img_gen_assignment)
    return await img_gen_single_image(img_gen_assignment)



import base64
from abc import ABC
from typing import Union

from pydantic import BaseModel
from typing_extensions import override

from pipelex.tools.misc.attribute_utils import AttributePolisher
from pipelex.tools.misc.filetype_utils import (
    FileType,
    detect_file_type_from_base64,
    detect_file_type_from_bytes,
    detect_file_type_from_path,
)
from pipelex.tools.typing.pydantic_utils import CustomBaseModel


class PromptImageTypedBase64(CustomBaseModel):
    base_64: bytes
    file_type: FileType


PromptImageTypedUrlOrBase64 = Union[str, PromptImageTypedBase64]


class PromptImage(BaseModel, ABC):
    pass


class PromptImagePath(PromptImage):
    file_path: str

    def get_file_type(self) -> FileType:
        return detect_file_type_from_path(self.file_path)

    def get_mime_type(self) -> str:
        return self.get_file_type().mime

    @override
    def __str__(self) -> str:
        return f"PromptImagePath(file_path='{self.file_path}')"


class PromptImageUrl(PromptImage):
    url: str

    @override
    def __str__(self) -> str:
        truncated_url = AttributePolisher.get_truncated_value(name="url", value=self.url)
        return f"PromptImageUrl(url='{truncated_url!r}')"

    @override
    def __format__(self, format_spec: str) -> str:
        return self.__str__()


class PromptImageBase64(PromptImage):
    base_64: bytes

    def get_file_type(self) -> FileType:
        return detect_file_type_from_base64(self.base_64)

    def get_mime_type(self) -> str:
        return self.get_file_type().mime

    def get_decoded_bytes(self) -> bytes:
        return base64.b64decode(self.base_64)

    @override
    def __str__(self) -> str:
        base_64_str = str(self.base_64)
        truncated_base_64 = AttributePolisher.get_truncated_value(name="base_64", value=base_64_str)
        return f"PromptImageBase64(base_64={truncated_base_64!r})"

    @override
    def __repr__(self) -> str:
        return self.__str__()

    @override
    def __format__(self, format_spec: str) -> str:
        return self.__str__()

    def make_prompt_image_typed_base64(self) -> PromptImageTypedBase64:
        return PromptImageTypedBase64(base_64=self.base_64, file_type=self.get_file_type())


class PromptImageBinary(PromptImage):
    binary: bytes

    def get_file_type(self) -> FileType:
        return detect_file_type_from_bytes(self.binary)

    def get_mime_type(self) -> str:
        return self.get_file_type().mime

    @override
    def __str__(self) -> str:
        return "PromptImageBinary(binary=...)"

    @override
    def __repr__(self) -> str:
        return self.__str__()



PLACEHOLDER_PREFIX = "placeholder"


def make_placeholder_value(key: str) -> str:
    """Create a placeholder value for a given key.

    Args:
        key: The key/variable name to create a placeholder for

    Returns:
        A placeholder value in the format "placeholder-for-{key}"

    """
    return f"{PLACEHOLDER_PREFIX}-for-{key}"


def value_is_placeholder(value: str | None) -> bool:
    """Check if a value is a placeholder based on the prefix.

    Args:
        value: The value to check

    Returns:
        True if the value starts with the placeholder prefix, False otherwise

    """
    return value is not None and value.startswith(PLACEHOLDER_PREFIX)



from typing import Literal

from pydantic import Field
from typing_extensions import override

from pipelex.cogt.img_gen.img_gen_job_components import AspectRatio, Background, OutputFormat
from pipelex.cogt.img_gen.img_gen_setting import ImgGenModelChoice
from pipelex.core.pipes.pipe_blueprint import PipeBlueprint


class PipeImgGenBlueprint(PipeBlueprint):
    type: Literal["PipeImgGen"] = "PipeImgGen"
    pipe_category: Literal["PipeOperator"] = "PipeOperator"
    img_gen_prompt: str | None = None
    img_gen_prompt_var_name: str | None = None

    model: ImgGenModelChoice | None = None

    # One-time settings (not in ImgGenSetting)
    aspect_ratio: AspectRatio | None = Field(default=None, strict=False)
    is_raw: bool | None = None
    seed: int | Literal["auto"] | None = None
    background: Background | None = Field(default=None, strict=False)
    output_format: OutputFormat | None = Field(default=None, strict=False)

    @override
    def validate_inputs(self):
        # check that we have either an img_gen_prompt passed as attribute or as a single text input
        if not self.inputs:
            if not self.img_gen_prompt:
                msg = "If no inputs are provided, you must provide an 'img_gen_prompt' as attribute."
                raise ValueError(msg)

        if self.inputs and self.img_gen_prompt:
            msg = "You must provide either an 'img_gen_prompt' as attribute or as a single text input, but not both"
            raise ValueError(msg)

        nb_inputs = self.nb_inputs
        if nb_inputs > 1:
            msg = f"Too many inputs provided for PipeImgGen: {self.input_names}. Only one input is allowed."
            raise ValueError(msg)

    @override
    def validate_output(self):
        pass



from pipelex.base_exceptions import PipelexError


class StuffArtefactError(PipelexError):
    pass


class StuffArtefactReservedFieldError(StuffArtefactError):
    pass


class StuffError(PipelexError):
    pass


class StuffContentTypeError(StuffError):
    def __init__(self, message: str, expected_type: str, actual_type: str):
        self.expected_type = expected_type
        self.actual_type = actual_type
        super().__init__(message)


class StuffContentValidationError(StuffError):
    """Raised when content validation fails during type conversion."""

    def __init__(self, original_type: str, target_type: str, validation_error: str):
        self.original_type = original_type
        self.target_type = target_type
        self.validation_error = validation_error
        super().__init__(f"Failed to validate content from {original_type} to {target_type}: {validation_error}")



from pydantic import BaseModel

from pipelex import log
from pipelex.core.memory.exceptions import WorkingMemoryStuffNotFoundError
from pipelex.core.memory.working_memory import WorkingMemory
from pipelex.core.pipes.inputs.exceptions import PipeInputError, PipeInputNotFoundError
from pipelex.core.pipes.pipe_factory import PipeFactory
from pipelex.core.pipes.pipe_output import PipeOutput
from pipelex.core.pipes.variable_multiplicity import VariableMultiplicity
from pipelex.core.stuffs.list_content import ListContent
from pipelex.hub import get_pipeline_tracker, get_required_pipe
from pipelex.pipe_controllers.batch.pipe_batch import PipeBatch
from pipelex.pipe_controllers.batch.pipe_batch_blueprint import PipeBatchBlueprint
from pipelex.pipe_controllers.condition.pipe_condition import PipeCondition
from pipelex.pipe_run.pipe_run_params import BatchParams, PipeRunParams
from pipelex.pipeline.job_metadata import JobMetadata


class SubPipe(BaseModel):
    pipe_code: str
    output_name: str | None = None
    output_multiplicity: VariableMultiplicity | None = None
    batch_params: BatchParams | None = None

    async def run_pipe(
        self,
        calling_pipe_code: str,
        working_memory: WorkingMemory,
        job_metadata: JobMetadata,
        sub_pipe_run_params: PipeRunParams,
    ) -> PipeOutput:
        """Run or dry run a single operation self."""
        if self.output_multiplicity:
            sub_pipe_run_params.output_multiplicity = self.output_multiplicity
        sub_pipe_run_params.batch_params = self.batch_params

        sub_pipe = get_required_pipe(pipe_code=self.pipe_code)

        # Case 1: Batch processing
        if batch_params := self.batch_params:
            try:
                working_memory.get_typed_object_or_attribute(name=batch_params.input_list_stuff_name, wanted_type=ListContent)
            except WorkingMemoryStuffNotFoundError as exc:
                msg = (
                    f"Input list stuff named '{batch_params.input_list_stuff_name}' required by sub_pipe '{self.pipe_code}' "
                    f"of pipe '{calling_pipe_code}' not found in working memory: {exc}"
                )
                raise PipeInputError(
                    message=msg, pipe_code=self.pipe_code, variable_name=batch_params.input_list_stuff_name, concept_code=None
                ) from exc

            try:
                item_stuff_requirement = sub_pipe.inputs.get_required_input_requirement(variable_name=batch_params.input_item_stuff_name)
            except PipeInputNotFoundError as exc:
                msg = (
                    f"Batch input item named '{batch_params.input_item_stuff_name}' from '{calling_pipe_code}' is not "
                    f"in SubPipe '{self.pipe_code}' input requirements: {sub_pipe.inputs}"
                )
                raise PipeInputError(
                    message=msg, pipe_code=self.pipe_code, variable_name=batch_params.input_item_stuff_name, concept_code=None
                ) from exc
            pipe_batch_blueprint = PipeBatchBlueprint(
                description=f"Batch processing for {self.pipe_code}",
                branch_pipe_code=self.pipe_code,
                output=sub_pipe.output.code,
                input_list_name=batch_params.input_list_stuff_name,
                input_item_name=batch_params.input_item_stuff_name,
                inputs={
                    batch_params.input_list_stuff_name: item_stuff_requirement.concept.concept_string,
                },
            )

            pipe_batch = PipeFactory[PipeBatch].make_from_blueprint(
                domain_code=sub_pipe.domain,
                pipe_code=self.pipe_code,
                blueprint=pipe_batch_blueprint,
                concept_codes_from_the_same_domain=[concept.code for concept in sub_pipe.concept_dependencies],
            )
            pipe_output = await pipe_batch.run_pipe(
                job_metadata=job_metadata,
                working_memory=working_memory,
                pipe_run_params=sub_pipe_run_params,
                output_name=self.output_name,
            )
        # Case 2: Condition processing
        elif isinstance(sub_pipe, PipeCondition):
            pipe_output = await sub_pipe.run_pipe(
                job_metadata=job_metadata,
                working_memory=working_memory,
                pipe_run_params=sub_pipe_run_params,
                output_name=self.output_name,
            )
        else:
            # Case 3: Normal processing
            required_variables = sub_pipe.required_variables()
            required_stuff_names = {req_var for req_var in required_variables if not req_var.startswith("_")}
            try:
                required_stuffs = working_memory.get_stuffs(names=required_stuff_names)
            except WorkingMemoryStuffNotFoundError as exc:
                sub_pipe_path = [*sub_pipe_run_params.pipe_stack, self.pipe_code]
                sub_pipe_path_str = ".".join(sub_pipe_path)
                error_details = f"SubPipe '{sub_pipe_path_str}', required_variables: {required_variables}, missing: '{exc.variable_name}'"
                msg = f"Some required stuff(s) not found: {error_details}"
                raise PipeInputError(message=msg, pipe_code=self.pipe_code, variable_name=exc.variable_name, concept_code=None) from exc
            log.verbose(required_stuffs, title=f"Required stuffs for {self.pipe_code}")
            pipe_output = await sub_pipe.run_pipe(
                job_metadata=job_metadata,
                working_memory=working_memory,
                pipe_run_params=sub_pipe_run_params,
                output_name=self.output_name,
            )
            if new_output_stuff := pipe_output.working_memory.get_optional_main_stuff():
                for stuff in required_stuffs:
                    get_pipeline_tracker().add_pipe_step(
                        from_stuff=stuff,
                        to_stuff=new_output_stuff,
                        pipe_code=self.pipe_code,
                        pipe_layer=sub_pipe_run_params.pipe_layers,
                        comment="SubPipe on required_stuff",
                    )
        return pipe_output



from typing import Any

from pydantic import Field, field_validator, model_validator
from pydantic.json_schema import SkipJsonSchema
from rich.console import Group
from rich.table import Table
from rich.text import Text
from typing_extensions import override

from pipelex.core.pipes.pipe_blueprint import PipeCategory, PipeType
from pipelex.core.stuffs.structured_content import StructuredContent
from pipelex.tools.misc.pretty import PrettyPrintable


class PipeSignature(StructuredContent):
    """PipeSignature is a contract for a pipe.

    It defines the inputs, outputs, and the purpose of the pipe without implementation details.

    Multiplicity Notation:
        Use bracket notation to specify multiplicity for both inputs and outputs:
        - No brackets: single item (default)
        - []: variable-length list
        - [N]: exactly N items (where N is a positive integer)

    Examples:
        - output = "Text" - one text items
        - output = "Text[]" - multiple text items
        - output = "Image[3]" - exactly 3 images
    """

    code: str = Field(description="Pipe code identifying the pipe. Must be snake_case.")
    type: PipeType | str = Field(description="Pipe type.")
    pipe_category: SkipJsonSchema[PipeCategory] = Field(description="Pipe category set according to its type.")
    description: str = Field(description="What the pipe does")
    inputs: dict[str, str] = Field(
        description=(
            "Input specifications mapping variable names to concept codes. "
            "Keys: input variable names in snake_case. "
            "Values: ConceptCodes in PascalCase. Don't use multiplicity brackets. "
        )
    )
    result: str = Field(description="Variable name for the pipe's result in snake_case. This name can be referenced as input in subsequent pipes.")
    output: str = Field(
        description=(
            "Output concept code in PascalCase with optional multiplicity brackets. "
            "Examples: 'Text' (single text), 'Article[]' (list of articles), 'Image[5]' (exactly 5 images)."
        )
    )
    pipe_dependencies: list[str] = Field(description="List of pipe codes that this pipe depends on. This is for the PipeControllers")

    @model_validator(mode="before")
    @classmethod
    def set_pipe_category(cls, values: dict[str, Any]) -> dict[str, Any]:
        try:
            type_str = values["type"]
        except TypeError as exc:
            msg = f"Invalid type for '{values}': could not get subscript, required for 'type'"
            raise ValueError(msg) from exc
        # we need to convert the type string to the PipeType enum because it arrives as a str implictly converted to enum but not yet
        the_type = PipeType(type_str)
        values["pipe_category"] = the_type.category
        return values

    @field_validator("type", mode="before")
    @classmethod
    def validate_type(cls, type_value: str) -> PipeType:
        return PipeType(type_value)

    @override
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        pipe_group = Group()
        if title:
            pipe_group.renderables.append(Text(title, style="bold"))
        pipe_group.renderables.append(Text.from_markup(f"Pipe Signature: [red]{self.code}[/red]\n", style="bold"))
        pipe_type = self.type.value if isinstance(self.type, PipeType) else str(self.type)
        pipe_group.renderables.append(Text.from_markup(f"Type: [bold magenta]{pipe_type}[/bold magenta] ({self.pipe_category.value})\n"))
        pipe_group.renderables.append(Text.from_markup(f"Description: [yellow italic]{self.description}[/yellow italic]\n"))

        # Create inputs section
        if not self.inputs:
            pipe_group.renderables.append(Text.from_markup("\nNo inputs"))
        elif len(self.inputs) == 1:
            # Single input: display as a simple line of text
            input_name, concept_spec = next(iter(self.inputs.items()))
            pipe_group.renderables.append(Text.from_markup(f"\nInput: [cyan]{input_name}[/cyan] ([bold green]{concept_spec}[/bold green])"))
        else:
            # Multiple inputs: display as a table
            inputs_table = Table(
                title="Inputs:",
                title_justify="left",
                title_style="not italic",
                show_header=False,
                show_edge=True,
                show_lines=True,
                border_style="dim",
            )
            inputs_table.add_column("Variable Name", style="cyan")
            inputs_table.add_column("Concept", style="bold green")
            for input_name, concept_spec in self.inputs.items():
                inputs_table.add_row(input_name, concept_spec)
            pipe_group.renderables.append(inputs_table)

        # Show output and result
        pipe_group.renderables.append(Text.from_markup(f"\nOutput concept: [bold green]{self.output}[/bold green]"))
        pipe_group.renderables.append(Text.from_markup(f"\nOutput name: [cyan]{self.result}[/cyan]"))

        # Show dependencies if any
        if self.pipe_dependencies:
            pipe_group.renderables.append(Text.from_markup(f"\nDependencies: [red]{', '.join(self.pipe_dependencies)}[/red]"))

        return pipe_group



import json
from typing import Any

import markdown
from typing_extensions import override

from pipelex import pretty_print_md
from pipelex.core.stuffs.stuff_content import StuffContent
from pipelex.tools.misc.file_utils import ensure_directory_exists, save_text_to_path


class TextContent(StuffContent):
    text: str

    @override
    def smart_dump(self) -> str | dict[str, Any] | list[str] | list[dict[str, Any]]:
        return self.text

    @property
    @override
    def short_desc(self) -> str:
        return f"some text ({len(self.text)} chars)"

    @override
    def __str__(self) -> str:
        return self.text

    @override
    def rendered_plain(self) -> str:
        return self.text

    @override
    def rendered_html(self) -> str:
        # Convert a markdown string to HTML and return HTML as a Unicode string.
        return markdown.markdown(self.text)

    @override
    def rendered_markdown(self, level: int = 1, is_pretty: bool = False) -> str:
        return self.text

    @override
    def rendered_json(self) -> str:
        return json.dumps({"text": self.text})

    def save_to_directory(self, directory: str):
        ensure_directory_exists(directory)
        filename = "text_content.txt"
        save_text_to_path(text=self.text, path=f"{directory}/{filename}")

    @override
    def pretty_print_content(self, title: str | None = None) -> None:
        pretty_print_md(self.text, title=title)



import logging
from typing import Any

from pipelex.tools.log.log_config import LogConfig, LogMode
from pipelex.tools.log.log_dispatch import LogDispatch
from pipelex.tools.log.log_formatter import EmojiLogFormatter, LevelAndEmojiLogFormatter
from pipelex.tools.log.log_levels import LOGGING_LEVEL_DEV, LOGGING_LEVEL_OFF, LOGGING_LEVEL_VERBOSE, LogLevel


class Log:
    """A class for managing logging configurations and operations."""

    ########################################################
    # Init and Configure
    ########################################################

    def __init__(self):
        """Initialize the Log class with default attributes."""
        self._log_config_instance: LogConfig | None = None
        self.rich_handler: logging.Handler | None = None
        self.poor_handler: logging.Handler | None = None
        self.log_dispatch: LogDispatch = LogDispatch()

    def set_log_mode(self, mode: LogMode):
        self.log_dispatch.set_log_mode(mode=mode)

    @property
    def _log_config(self) -> LogConfig:
        """Get the log configuration, raising an error if it's not set.

        Returns:
            LogConfig: The current log configuration.

        Raises:
            RuntimeError: If the log configuration is not set.

        """
        if self._log_config_instance is None:
            msg = "LogConfig is not set. You must initialize Pipelex first."
            raise RuntimeError(msg)
        return self._log_config_instance

    def reset(self):
        """Reset the logging system."""
        # Remove all handlers from the root logger
        root_logger = logging.getLogger()
        for handler in root_logger.handlers[:]:
            root_logger.removeHandler(handler)
            handler.close()

        # Remove handlers from poor loggers
        if self._log_config_instance:
            poor_loggers = {
                *self._log_config_instance.poor_loggers,
                self._log_config_instance.generic_poor_logger,
            }
            for logger_name in poor_loggers:
                logger = logging.getLogger(logger_name)
                for handler in logger.handlers[:]:
                    logger.removeHandler(handler)
                    handler.close()

        logging.shutdown()
        self._log_config_instance = None
        self.rich_handler = None
        self.poor_handler = None
        self.log_dispatch.reset()

    def configure(self, log_config: LogConfig):
        """Configure the logging system with the given project name and log configuration.

        Args:
            log_config (LogConfig): The log configuration to use.

        Raises:
            RuntimeError: If the log configuration is already set.

        """
        if self._log_config_instance is not None:
            msg = "LogConfig is already set. You can only call log.configure() once."
            raise RuntimeError(msg)

        self.log_dispatch.configure(log_config=log_config)

        self._log_config_instance = log_config

        self.rich_handler = log_config.rich_log_config.make_rich_handler(target=log_config.console_log_target)
        self.rich_handler.setFormatter(EmojiLogFormatter())

        # Configure the root logger
        root_logger = logging.getLogger()
        root_logger.setLevel(log_config.default_log_level.int_logging_level)
        root_logger.addHandler(self.rich_handler)

        self.poor_handler = logging.StreamHandler()
        self.poor_handler.setFormatter(LevelAndEmojiLogFormatter())
        poor_loggers = {*log_config.poor_loggers, log_config.generic_poor_logger}
        for logger_name in poor_loggers:
            logger = logging.getLogger(logger_name)
            logger.setLevel(log_config.default_log_level.int_logging_level)
            logger.addHandler(self.poor_handler)
            logger.propagate = False

        self.set_levels_for_packages(package_log_levels=log_config.package_log_levels)

        self.verbose("Logs configured and config set")

    def set_poor_log_formatter(self, formatter: logging.Formatter):
        """Set the formatter for the poor log handler.

        Args:
            formatter (logging.Formatter): The formatter to use for poor logging.

        """
        if self.poor_handler is None:
            msg = "Poor log handler is not set."
            raise RuntimeError(msg)
        self.poor_handler.setFormatter(formatter)

    def _should_ignore(self, problem_id: str | None = None) -> bool:
        """Check if a log message should be ignored based on the problem ID.

        Args:
            problem_id (str | None): The problem ID to check.

        Returns:
            bool: True if the message should be ignored, False otherwise.

        """
        return bool(problem_id) and problem_id in self._log_config.silenced_problem_ids

    ########################################################
    # Public methods
    ########################################################

    def set_level_by_int(self, level_int: int):
        """Set the log level using an integer value.

        Args:
            level_int (int): The integer representation of the log level.

        """
        logging.getLogger().setLevel(level_int)

    def set_level_by_name(self, level_name: str):
        """Set the log level using a string name.

        Args:
            level_name (str): The name of the log level.

        """
        if level_name.upper() == LogLevel.DEV:
            level = LOGGING_LEVEL_DEV
        elif level_name.upper() == LogLevel.OFF:
            level = LOGGING_LEVEL_OFF
        else:
            level = getattr(logging, level_name.upper())
        self.set_level_by_int(level)

    def set_level(self, level: LogLevel):
        """Set the default log level for all loggers.

        Args:
            level (LogLevel): The log level to set.

        """
        self.set_level_by_int(level_int=level.int_logging_level)

    def set_level_for_package(self, package_name: str, level: LogLevel):
        """Set the log level for a specific package.

        Args:
            package_name (str): The name of the package.
            level (LogLevel): The log level to set for the package.

        """
        logger_name = package_name.replace("-", ".")
        logging.getLogger(logger_name).setLevel(level.int_logging_level)

    def set_levels_for_packages(self, package_log_levels: dict[str, LogLevel]):
        """Set log levels for multiple packages.

        Args:
            package_log_levels (Dict[str, LogLevel]): A dictionary mapping package names to log levels.

        """
        for package_name, level in package_log_levels.items():
            self.set_level_for_package(package_name=package_name, level=level)

    def verbose(
        self,
        content: str | Any,
        title: str | None = None,
        inline: str | None = None,
    ):
        """Log a verbose message.

        Args:
            content (Union[str, Any]): The content to log.
            title (str | None, optional): The title of the log message. Defaults to None.
            inline (str | None, optional): Inline title for the log message. Defaults to None.
                Used to display the title inline, only if the title arg is None.

        """
        severity = LOGGING_LEVEL_VERBOSE
        self.log_dispatch.dispatch(content=content, severity=severity, title=title, inline=inline)

    def debug(
        self,
        content: str | Any,
        title: str | None = None,
        inline: str | None = None,
    ):
        """Log a debug message.

        Args:
            content (Union[str, Any]): The content to log.
            title (str | None, optional): The title of the log message. Defaults to None.
            inline (str | None, optional): Inline title for the log message. Defaults to None.
                Used to display the title inline, only if the title arg is None.

        """
        severity = logging.DEBUG
        self.log_dispatch.dispatch(content=content, severity=severity, title=title, inline=inline)

    def dev(
        self,
        content: str | Any,
        title: str | None = None,
        inline: str | None = None,
    ):
        """Log a development message.

        Args:
            content (Union[str, Any]): The content to log.
            title (str | None, optional): The title of the log message. Defaults to None.
            inline (str | None, optional): Inline title for the log message. Defaults to None.
                Used to display the title inline, only if the title arg is None.

        """
        severity = LOGGING_LEVEL_DEV
        self.log_dispatch.dispatch(content=content, severity=severity, title=title, inline=inline)

    def info(
        self,
        content: str | Any,
        title: str | None = None,
        inline: str | None = None,
    ):
        """Log an info message.

        Args:
            content (Union[str, Any]): The content to log.
            title (str | None, optional): The title of the log message. Defaults to None.
            inline (str | None, optional): Inline title for the log message. Defaults to None.
                Used to display the title inline, only if the title arg is None.

        """
        severity = logging.INFO
        self.log_dispatch.dispatch(content=content, severity=severity, title=title, inline=inline)

    def warning(
        self,
        content: str | Any,
        title: str | None = None,
        inline: str | None = None,
        problem_id: str | None = None,
    ):
        """Log a warning message.

        Args:
            content (Union[str, Any]): The content to log.
            title (str | None, optional): The title of the log message. Defaults to None.
            inline (str | None, optional): Inline title for the log message. Defaults to None.
                Used to display the title inline, only if the title arg is None.
            problem_id (str | None, optional): A problem ID to associate with the warning. Defaults to None.

        """
        if self._should_ignore(problem_id=problem_id):
            return
        severity = logging.WARNING
        self.log_dispatch.dispatch(content=content, severity=severity, title=title, inline=inline)

    def error(
        self,
        content: str | Any,
        title: str | None = None,
        inline: str | None = None,
        include_exception: bool = False,
        problem_id: str | None = None,
    ):
        """Log an error message.

        Args:
            content (Union[str, Any]): The content to log.
            title (str | None, optional): The title of the log message. Defaults to None.
            inline (str | None, optional): Inline title for the log message. Defaults to None.
                Used to display the title inline, only if the title arg is None.
            include_exception (bool, optional): Whether to include exception information. Defaults to False.
            problem_id (str | None, optional): A problem ID to associate with the error. Defaults to None.

        """
        if self._should_ignore(problem_id=problem_id):
            return
        severity = logging.ERROR
        self.log_dispatch.dispatch(
            content=content,
            severity=severity,
            title=title,
            inline=inline,
            include_exception=include_exception,
        )

    def critical(
        self,
        content: str | Any,
        title: str | None = None,
        inline: str | None = None,
        include_exception: bool = False,
        problem_id: str | None = None,
    ):
        """Log a critical message.

        Args:
            content (Union[str, Any]): The content to log.
            title (str | None, optional): The title of the log message. Defaults to None.
            inline (str | None, optional): Inline title for the log message. Defaults to None.
                Used to display the title inline, only if the title arg is None.
            include_exception (bool, optional): Whether to include exception information. Defaults to False.
            problem_id (str | None, optional): A problem ID to associate with the critical message. Defaults to None.

        """
        if self._should_ignore(problem_id=problem_id):
            return
        severity = logging.CRITICAL
        self.log_dispatch.dispatch(
            content=content,
            severity=severity,
            title=title,
            inline=inline,
            include_exception=include_exception,
        )


log = Log()



# pyright: reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownParameterType=false
# pyright: reportMissingTypeArgument=false
from typing import Any

import networkx as nx
import yaml

from pipelex import log
from pipelex.pipeline.track.exceptions import JobHistoryError
from pipelex.pipeline.track.tracker_config import TrackerConfig
from pipelex.pipeline.track.tracker_models import (
    EdgeAttributeKey,
    EdgeCategory,
    GraphTree,
    NodeAttributeKey,
    NodeCategory,
    SubGraphClassDef,
)
from pipelex.tools.misc.mermaid_utils import clean_str_for_mermaid_node_title, make_mermaid_url
from pipelex.tools.misc.string_utils import snake_to_capitalize_first_letter


def nice_edge_tag(edge_tag: str) -> str:
    return f'"{snake_to_capitalize_first_letter(edge_tag)}"'


def _indent_line(line: str, indent: int) -> str:
    return f"{'    ' * indent}{line}"


class PipelineFlowChart:
    def __init__(self, nx_graph: nx.DiGraph, start_node: str, tracker_config: TrackerConfig):
        self.nx_graph = nx_graph
        self._tracker_config = tracker_config
        self.is_debug_mode = tracker_config.is_debug_mode
        self.start_node = start_node
        self.sub_graph_class_defs: list[SubGraphClassDef] = []
        for sub_graph_index, sub_graph_color in enumerate(self._tracker_config.sub_graph_colors):
            class_def_letter = chr(ord("a") + sub_graph_index)
            class_def_name = f"sub_{class_def_letter}"
            self.sub_graph_class_defs.append(SubGraphClassDef(name=class_def_name, color=sub_graph_color))

    def generate_mermaid_flowchart(
        self,
        title: str | None = None,
        subtitle: str | None = None,
    ) -> tuple[str, str]:
        nb_nodes = len(self.nx_graph.nodes)
        if nb_nodes == 0:
            msg = "Graph has no nodes"
            raise JobHistoryError(msg)
        log.verbose(f"Generating mermaid flowchart for the whole graph which holds {nb_nodes} nodes")
        mermaid_settings: dict[str, Any] = {}
        if title:
            mermaid_settings["title"] = title
        mermaid_settings["config"] = {}
        if self._tracker_config.applied_theme:
            mermaid_settings["config"]["theme"] = self._tracker_config.applied_theme
        if self._tracker_config.applied_layout:
            mermaid_settings["config"]["layout"] = self._tracker_config.applied_layout
        if self._tracker_config.applied_wrapping_width:
            mermaid_settings["config"]["flowchart"] = {"wrappingWidth": self._tracker_config.applied_wrapping_width}
        mermaid_code = "---\n"
        mermaid_code += yaml.dump(mermaid_settings)
        mermaid_code += "---\n"
        mermaid_code += "flowchart LR\n"

        graph_tree: GraphTree = GraphTree(nodes_by_subgraph={})

        # First pass: Collect nodes into their respective subgraphs
        for node in self.nx_graph.nodes:
            node_attributes = self.nx_graph.nodes[node]
            if not node_attributes:
                msg = f"Node attributes are empty for node '{node}'"
                raise JobHistoryError(msg)
            node_pipe_layer = node_attributes.get(NodeAttributeKey.SUBGRAPH)
            if not node_pipe_layer:
                node_pipe_layer = "Unknown"
            elif not isinstance(node_pipe_layer, str):
                msg = f"Node '{node}' has no pipe stack: {node_attributes}"
                raise JobHistoryError(msg)

            sub_graph = node_pipe_layer or "root"
            # Split sub_graph by "-" and keep the last token
            if "-" in sub_graph:
                sub_graph = sub_graph.split("-")[-1]
            if sub_graph not in graph_tree.nodes_by_subgraph:
                graph_tree.nodes_by_subgraph[sub_graph] = []
            graph_tree.nodes_by_subgraph[sub_graph].append(node)

        log.verbose(graph_tree, title="Graph tree")

        subgraph_lines = self.generate_subgraph_lines(graph_tree)
        mermaid_code += "\n".join(subgraph_lines)
        mermaid_code += "\n"

        # Generate subtitle
        if subtitle:
            # this is a hack to add something that looks like a subtitle, but it's actually a node with no stroke and no visible link
            # if self.start_node is None:
            #     msg = "Start node is not set"
            #     raise JobHistoryError(msg)
            mermaid_code += f"""
    classDef subtitleNodeClass fill:transparent,stroke:#333,stroke-width:0px;
    __subtitle__["{subtitle}"]
    class __subtitle__ subtitleNodeClass;
    __subtitle__ --> {self.start_node}
    linkStyle 0 stroke:transparent,stroke-width:0px
"""

        for sub_graph_class_def in self.sub_graph_class_defs:
            mermaid_code += f"""
    classDef {sub_graph_class_def.name} fill:{sub_graph_class_def.color},color:#333,stroke:#333;
"""

        # Generate edges
        for edge in self.nx_graph.edges(data=True):
            source, target, edge_data = edge
            edge_tag: str
            edge_type = EdgeCategory(edge_data[EdgeAttributeKey.EDGE_CATEGORY])
            match edge_type:
                case EdgeCategory.PIPE:
                    if pipe_code := edge_data.get(EdgeAttributeKey.PIPE_CODE):
                        edge_tag = nice_edge_tag(pipe_code)
                        mermaid_code += f"    {source} -- {edge_tag} {self._tracker_config.pipe_edge_style} {target}\n"
                    else:
                        msg = f"Pipe edge missing pipe code: {edge_data}"
                        raise JobHistoryError(msg)
                case EdgeCategory.BATCH:
                    mermaid_code += f"    {source} {self._tracker_config.branch_edge_style} {target}\n"
                case EdgeCategory.AGGREGATE:
                    mermaid_code += f"    {source} {self._tracker_config.aggregate_edge_style} {target}\n"
                case EdgeCategory.CONDITION:
                    condition_expression = edge_data.get(EdgeAttributeKey.CONDITION_EXPRESSION)
                    if not condition_expression:
                        msg = f"Condition edge missing condition expression: {edge_data}"
                        raise JobHistoryError(msg)
                    edge_tag = nice_edge_tag(condition_expression)
                    mermaid_code += f"    {source} {self._tracker_config.condition_edge_style} {target}\n"
                case EdgeCategory.CHOICE:
                    chosen_pipe_code = edge_data.get(EdgeAttributeKey.CHOSEN_PIPE)
                    if not chosen_pipe_code:
                        msg = f"No chosen pipe code set for edge {source} --- {target}"
                        raise JobHistoryError(msg)
                    edge_tag = nice_edge_tag(chosen_pipe_code)
                    mermaid_code += f"    {source} -- {edge_tag} {self._tracker_config.choice_edge_style} {target}\n"

        url = make_mermaid_url(mermaid_code)
        return mermaid_code, url

    def generate_subgraph_lines(self, graph_tree: GraphTree) -> list[str]:
        subgraph_lines: list[str] = []
        subgraph_class_lines: list[str] = []

        for cycle, (subgraph_name, nodes) in enumerate(graph_tree.nodes_by_subgraph.items()):
            node_lines: list[str] = []
            for node in nodes:
                # log.verbose(f"generate_subgraph_lines for node '{node}'")
                node_attributes = self.nx_graph.nodes[node]
                if not node_attributes:
                    msg = f"Node attributes are empty for node '{node}'"
                    raise JobHistoryError(msg)
                node_category = NodeCategory(node_attributes[NodeAttributeKey.CATEGORY])
                node_tag = node_attributes[NodeAttributeKey.TAG]
                if not node_tag:
                    msg = f"Node tag is empty for node '{node}'"
                    raise JobHistoryError(msg)
                node_text = node_tag
                if self.is_debug_mode:
                    if node_comment := node_attributes.get(NodeAttributeKey.COMMENT):
                        node_text += f"\n\n{node_comment}"
                    else:
                        node_text += "\n\nNo comment"
                    if node_debug_info := node_attributes.get(NodeAttributeKey.DEBUG_INFO):
                        node_text += f"\n\n{node_debug_info}"
                match node_category:
                    case NodeCategory.SPECIAL:
                        node_lines.append(f'{node}(["{node_text}"])')
                    case NodeCategory.STUFF:
                        node_lines.append(f'{node}["{node_text}"]')
                    case NodeCategory.CONDITION:
                        node_lines.append(f'{node}{{"{node_text}"}}')
                if node_description := node_attributes.get(NodeAttributeKey.DESCRIPTION) and self._tracker_config.is_include_interactivity:
                    if not isinstance(node_description, str):
                        msg = f"Node description is not a string: {node_description}"
                        raise JobHistoryError(msg)
                    node_description = clean_str_for_mermaid_node_title(node_description)
                    node_lines.append(f'click {node} stuff_node_callback "{node_description}"')

            if not node_lines:
                msg = f"No node lines found for subgraph '{subgraph_name}'"
                raise JobHistoryError(msg)

            if subgraph_name == "root":
                subgraph_lines.extend(_indent_line(mermaid_line, 2) for mermaid_line in node_lines)
            else:
                subgraph_lines.append(_indent_line(f'subgraph "{subgraph_name}"', 1))
                subgraph_lines.append(_indent_line("direction LR", 1))
                subgraph_lines.extend(_indent_line(mermaid_line, 2) for mermaid_line in node_lines)
                subgraph_lines.append(_indent_line("end", 1))

                class_def = self.sub_graph_class_defs[cycle % len(self.sub_graph_class_defs)]
                subgraph_class_lines.append(f"class {subgraph_name} {class_def.name};")

        subgraph_lines.extend(subgraph_class_lines)

        return subgraph_lines



from itertools import groupby

from pydantic import RootModel
from rich import box
from rich.table import Table
from typing_extensions import override

from pipelex import pretty_print
from pipelex.core.pipes.pipe_abstract import PipeAbstract
from pipelex.libraries.pipe.exceptions import PipeLibraryError, PipeNotFoundError
from pipelex.libraries.pipe.pipe_library_abstract import PipeLibraryAbstract
from pipelex.types import Self

PipeLibraryRoot = dict[str, PipeAbstract]


class PipeLibrary(RootModel[PipeLibraryRoot], PipeLibraryAbstract):
    @override
    def setup(self):
        pass

    @override
    def teardown(self):
        self.root = {}

    @override
    def reset(self):
        self.teardown()
        self.setup()

    @classmethod
    def make_empty(cls) -> Self:
        library = cls(root=PipeLibraryRoot())
        library.setup()
        return library

    @override
    def add_new_pipe(self, pipe: PipeAbstract):
        if pipe.code in self.root:
            msg = (
                f"Pipe '{pipe.code}' already exists in the library. You might be running the same pipe twice in the same pipeline."
                "We do not yet handle this case, so please avoid running the same pipe twice in the same pipeline"
                "Or consider adding for good in the library and call it by its code."
            )
            raise PipeLibraryError(msg)
        self.root[pipe.code] = pipe

    @override
    def add_pipes(self, pipes: list[PipeAbstract]):
        for pipe in pipes:
            self.add_new_pipe(pipe=pipe)

    @override
    def get_optional_pipe(self, pipe_code: str) -> PipeAbstract | None:
        return self.root.get(pipe_code)

    @override
    def get_required_pipe(self, pipe_code: str) -> PipeAbstract:
        the_pipe = self.get_optional_pipe(pipe_code=pipe_code)
        if not the_pipe:
            msg = f"Pipe '{pipe_code}' not found. Check for typos and make sure it is declared in plx file in an imported package."
            raise PipeNotFoundError(msg)
        return the_pipe

    @override
    def get_pipes(self) -> list[PipeAbstract]:
        return list(self.root.values())

    @override
    def get_pipes_dict(self) -> dict[str, PipeAbstract]:
        return self.root

    @override
    def remove_pipes_by_codes(self, pipe_codes: list[str]) -> None:
        # TODO: We should create a separate library, that copies the original one, and then removes the pipes from it
        # Then run the dry run + validation to see if removing those pipe has not broken any other pipe.
        # If validated, it should update the real library.
        for pipe_code in pipe_codes:
            if pipe_code in self.root:
                del self.root[pipe_code]

    @override
    def pretty_list_pipes(self) -> None:
        def _format_concept_code(concept_code: str | None, current_domain: str) -> str:
            """Format concept code by removing domain prefix if it matches current domain."""
            if not concept_code:
                return ""
            parts = concept_code.split(".")
            if len(parts) == 2 and parts[0] == current_domain:
                return parts[1]
            return concept_code

        pipes = self.get_pipes()

        # Sort pipes by domain and code
        ordered_items = sorted(pipes, key=lambda pipe: (pipe.domain or "", pipe.code or ""))

        # Create dictionary for return value
        pipes_dict: dict[str, dict[str, dict[str, str]]] = {}

        # Group by domain and create separate tables
        for domain, domain_pipes in groupby(ordered_items, key=lambda pipe: pipe.domain):
            table = Table(
                title=f"[bold magenta]domain = {domain}[/]",
                show_header=True,
                show_lines=True,
                header_style="bold cyan",
                box=box.SQUARE_DOUBLE_HEAD,
                border_style="blue",
            )

            table.add_column("Code", style="green")
            table.add_column("Definition", style="white")
            table.add_column("Input", style="yellow")
            table.add_column("Output", style="yellow")

            pipes_dict[domain] = {}

            for pipe in domain_pipes:
                inputs = pipe.inputs
                formatted_inputs = [f"{name}: {_format_concept_code(requirement.concept.code, domain)}" for name, requirement in inputs.items]
                formatted_inputs_str = ", ".join(formatted_inputs)
                output_code = _format_concept_code(pipe.output.code, domain)

                table.add_row(
                    pipe.code,
                    pipe.description or "",
                    formatted_inputs_str,
                    output_code,
                )

                pipes_dict[domain][pipe.code] = {
                    "description": pipe.description or "",
                    "inputs": formatted_inputs_str,
                    "output": pipe.output.code,
                }

            pretty_print(table)



def strip_multiplicity_from_concept_string_or_code(concept_string_or_code: str) -> str:
    """Strip multiplicity from a concept string or code.

    Args:
        concept_string_or_code: The concept string or code to strip multiplicity from

    Returns:
        The concept string or code without multiplicity
    """
    if "[" in concept_string_or_code:
        return concept_string_or_code.split("[")[0]
    return concept_string_or_code



from jinja2 import meta
from jinja2.exceptions import (
    TemplateSyntaxError,
    UndefinedError,
)

from pipelex.cogt.templating.template_category import TemplateCategory
from pipelex.tools.jinja2.jinja2_environment import make_jinja2_env_without_loader
from pipelex.tools.jinja2.jinja2_errors import Jinja2DetectVariablesError, Jinja2StuffError


def detect_jinja2_required_variables(
    template_category: TemplateCategory,
    template_source: str,
) -> set[str]:
    """Returns a list of variables required by the Jinja2 template.

    Args:
        template_category: Category of the template (HTML, MARKDOWN, etc.), used to set the appropriate jinja2 environment settings
        template_source: Jinja2 template string

    Returns:
        List of variable names required by the template

    Raises:
        Jinja2DetectVariablesError: If there is an error parsing the template

    """
    jinja2_env = make_jinja2_env_without_loader(
        template_category=template_category,
    )

    try:
        parsed_ast = jinja2_env.parse(template_source)
        undeclared_variables = meta.find_undeclared_variables(parsed_ast)
    except Jinja2StuffError as stuff_error:
        msg = f"Jinja2 detect variables — stuff error: '{stuff_error}', template_category: {template_category}, template_source:\n{template_source}"
        raise Jinja2DetectVariablesError(msg) from stuff_error
    except TemplateSyntaxError as syntax_error:
        msg = f"Jinja2 detect variables — syntax error: '{syntax_error}', template_category: {template_category}, template_source:\n{template_source}"
        raise Jinja2DetectVariablesError(msg) from syntax_error
    except UndefinedError as undef_error:
        msg = (
            f"Jinja2 detect variables — undefined error: '{undef_error}', template_category: {template_category}, template_source:\n{template_source}"
        )
        raise Jinja2DetectVariablesError(msg) from undef_error

    return undeclared_variables



"""Types for the init command UI."""

from pipelex.types import StrEnum


class InitFocus(StrEnum):
    """Focus options for initialization."""

    ALL = "all"
    CONFIG = "config"
    INFERENCE = "inference"
    ROUTING = "routing"
    TELEMETRY = "telemetry"



import asyncio

from pipelex.client.protocol import PipelineInputs
from pipelex.core.memory.working_memory import WorkingMemory
from pipelex.core.pipes.pipe_output import PipeOutput
from pipelex.hub import get_pipe_router
from pipelex.pipe_run.pipe_run_mode import PipeRunMode
from pipelex.pipe_run.pipe_run_params import VariableMultiplicity
from pipelex.pipeline.pipeline_run_setup import pipeline_run_setup


async def start_pipeline(
    library_id: str | None = None,
    library_dirs: list[str] | None = None,
    pipe_code: str | None = None,
    plx_content: str | None = None,
    inputs: PipelineInputs | WorkingMemory | None = None,
    output_name: str | None = None,
    output_multiplicity: VariableMultiplicity | None = None,
    dynamic_output_concept_code: str | None = None,
    pipe_run_mode: PipeRunMode | None = None,
    search_domains: list[str] | None = None,
) -> tuple[str, asyncio.Task[PipeOutput]]:
    """Start a pipeline in the background.

    This function mirrors ``execute_pipeline`` but returns immediately with the
    ``pipeline_run_id`` and a task instead of waiting for the pipe run to complete.
    The actual execution is scheduled on the current event loop using
    ``asyncio.create_task``.

    Parameters
    ----------
    library_id:
        Unique identifier for the library instance. If not provided, defaults to the
        auto-generated ``pipeline_run_id``. Use a custom ID when you need to manage
        multiple library instances or maintain library state across executions.
    library_dirs:
        List of directory paths to load pipe definitions from. If not provided, loads
        from the current working directory (the directory from which the Python script
        is executed). Ignored when ``plx_content`` is provided.
    pipe_code:
        Code identifying the pipe to execute. Required when ``plx_content`` is not
        provided. When both ``plx_content`` and ``pipe_code`` are provided, the
        specified pipe from the PLX content will be executed (overriding any
        ``main_pipe`` defined in the content).
    plx_content:
        Complete PLX file content as a string. When provided, only this content is
        loaded into the library, creating an isolated execution environment. The pipe
        to execute is determined by ``pipe_code`` (if provided) or the ``main_pipe``
        property in the PLX content.
    inputs:
        Inputs passed to the pipeline. Can be either a ``PipelineInputs`` dictionary
        or a ``WorkingMemory`` instance.
    output_name:
        Name of the output slot to write to.
    output_multiplicity:
        Output multiplicity specification.
    dynamic_output_concept_code:
        Override the dynamic output concept code.
    pipe_run_mode:
        Pipe run mode: ``PipeRunMode.LIVE`` or ``PipeRunMode.DRY``. If not specified,
        inferred from the environment variable ``PIPELEX_FORCE_DRY_RUN_MODE``. Defaults
        to ``PipeRunMode.LIVE`` if the environment variable is not set.
    search_domains:
        List of domains to search for pipes. The executed pipe's domain is automatically
        added if not already present.

    Returns:
    -------
    tuple[str, asyncio.Task[PipeOutput]]
        The ``pipeline_run_id`` of the newly started pipeline and a task that
        can be awaited to get the pipe output.

    """
    pipe_job, pipeline_run_id, library_id = await pipeline_run_setup(
        library_id=library_id,
        library_dirs=library_dirs,
        pipe_code=pipe_code,
        plx_content=plx_content,
        inputs=inputs,
        output_name=output_name,
        output_multiplicity=output_multiplicity,
        dynamic_output_concept_code=dynamic_output_concept_code,
        pipe_run_mode=pipe_run_mode,
        search_domains=search_domains,
    )

    task: asyncio.Task[PipeOutput] = asyncio.create_task(get_pipe_router().run(pipe_job))

    return pipeline_run_id, task



from typing import Protocol, runtime_checkable

from pipelex.cogt.usage.token_category import NbTokensByCategoryDict
from pipelex.plugins.bedrock.bedrock_message import BedrockMessageDictList


@runtime_checkable
class BedrockClientProtocol(Protocol):
    async def chat(
        self,
        messages: BedrockMessageDictList,
        system_text: str | None,
        model: str,
        temperature: float,
        max_tokens: int | None = None,
    ) -> tuple[str, NbTokensByCategoryDict]: ...



from abc import ABC, abstractmethod

LATEST_SECRET_VERSION_NAME = "latest"


class SecretsProviderAbstract(ABC):
    @abstractmethod
    def get_required_secret(self, secret_id: str) -> str: ...

    @abstractmethod
    def get_optional_secret(self, secret_id: str) -> str | None: ...

    @abstractmethod
    def get_required_secret_specific_version(self, secret_id: str, version_id: str) -> str: ...

    @abstractmethod
    def get_optional_secret_specific_version(self, secret_id: str, version_id: str) -> str | None: ...

    @abstractmethod
    def set_secret_as_env_var(self, secret_id: str, version_id: str = LATEST_SECRET_VERSION_NAME): ...

    def get_secret(self, secret_id: str) -> str:
        return self.get_required_secret(secret_id=secret_id)



import shutil
from abc import ABC, abstractmethod
from io import StringIO
from typing import Any, ClassVar

from kajson import kajson
from rich.console import Console, Group
from rich.json import JSON
from rich.markdown import Markdown
from rich.measure import Measurement
from rich.panel import Panel
from rich.pretty import Pretty
from rich.style import StyleType
from rich.syntax import Syntax
from rich.table import Table
from rich.terminal_theme import TerminalTheme
from rich.text import Text, TextType

from pipelex.tools.misc.terminal_utils import BOLD_FONT, RESET_FONT, TerminalColor, print_to_stderr
from pipelex.types import StrEnum

TEXT_COLOR = TerminalColor.WHITE
TITLE_COLOR = TerminalColor.CYAN
BORDER_COLOR = TerminalColor.YELLOW

# TODO: Make PrettyPrinter a manager so we can init it with a proper config
PRETTY_WIDTH_MIN: int = 125
PRETTY_WIDTH_FOR_EXPORT: int = 100
MAX_RENDER_DEPTH = 6
EXPORT_THEME = TerminalTheme(
    (0, 0, 0),
    (197, 200, 198),
    [
        (75, 78, 85),
        (204, 85, 90),
        (152, 168, 75),
        (208, 179, 68),
        (96, 138, 177),
        (152, 114, 159),
        (104, 160, 179),
        (197, 200, 198),
        (154, 155, 153),
    ],
    [
        (255, 38, 39),
        (0, 130, 61),
        (208, 132, 66),
        (25, 132, 233),
        (255, 44, 122),
        (57, 130, 128),
        (253, 253, 197),
    ],
)

PrettyPrintable = Markdown | Text | JSON | Table | Group | Syntax | Pretty


class PrettyRenderable(ABC):
    @abstractmethod
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        pass


class PrettyPrintMode(StrEnum):
    RICH = "rich"
    POOR = "poor"


def pretty_print(
    content: str | Any,
    title: TextType | None = None,
    subtitle: TextType | None = None,
    inner_title: str | None = None,
    border_style: StyleType | None = None,
    width: int | None = None,
    console_width: int | None = None,
):
    PrettyPrinter.pretty_print(
        content=content,
        title=title,
        subtitle=subtitle,
        inner_title=inner_title,
        border_style=border_style,
        width=width,
        console_width=console_width,
    )


def pretty_print_md(
    content: str,
    title: TextType | None = None,
    subtitle: TextType | None = None,
    inner_title: str | None = None,
    border_style: StyleType | None = None,
    width: int | None = None,
    console_width: int | None = None,
):
    width = width or PrettyPrinter.pretty_width()
    md_content = Markdown(content)
    PrettyPrinter.pretty_print(
        content=md_content,
        title=title,
        subtitle=subtitle,
        inner_title=inner_title,
        border_style=border_style,
        width=width,
        console_width=console_width,
    )


class PrettyPrinter:
    mode: ClassVar[PrettyPrintMode] = PrettyPrintMode.RICH

    @classmethod
    def pretty_print(
        cls,
        content: str | Any,
        title: TextType | None = None,
        subtitle: TextType | None = None,
        inner_title: str | None = None,
        border_style: StyleType | None = None,
        width: int | None = None,
        console_width: int | None = None,
    ):
        match cls.mode:
            case PrettyPrintMode.RICH:
                cls.pretty_print_using_rich(
                    content=content,
                    title=title,
                    subtitle=subtitle,
                    inner_title=inner_title,
                    border_style=border_style,
                    width=width,
                    console_width=console_width,
                )
            case PrettyPrintMode.POOR:
                cls.pretty_print_without_rich(content=content, title=title, subtitle=subtitle, inner_title=inner_title, console_width=console_width)

    @classmethod
    def pretty_print_using_rich(
        cls,
        content: str | Any,
        title: TextType | None = None,
        subtitle: TextType | None = None,
        inner_title: str | None = None,
        border_style: StyleType | None = None,
        width: int | None = None,
        console_width: int | None = None,
    ):
        panel = cls.make_pretty_panel(
            content=content,
            title=title,
            subtitle=subtitle,
            inner_title=inner_title,
            border_style=border_style,
            width=width,
            console_width=console_width,
        )

        Console(width=console_width).print("", panel, "", sep="\n")

    @classmethod
    def pretty_width(cls, width: int | None = None, depth: int | None = None) -> int:
        terminal_width = shutil.get_terminal_size().columns
        absolute_width = width or min(max(PRETTY_WIDTH_MIN, terminal_width // 2), terminal_width)
        if depth is not None:
            # Calculate adaptive width factor based on depth to prevent excessive narrowing
            # Factor decreases slowly: depth 0->1.0, depth 1->0.9, depth 2->0.8, etc., min 0.5
            width_factor = max(0.5, 1.0 - (depth * 0.1))
            return int(absolute_width * width_factor)
        else:
            return absolute_width

    @classmethod
    def make_pretty_panel(
        cls,
        content: str | Any,
        title: TextType | None = None,
        subtitle: TextType | None = None,
        inner_title: str | None = None,
        border_style: StyleType | None = None,
        width: int | None = None,
        console_width: int | None = None,
    ) -> Panel:
        pretty = cls.make_pretty(content, inner_title=inner_title, depth=0)
        # When width is not specified, measure the content to determine optimal console width
        if width is None:
            # Create a console to measure the panel
            measure_console = Console(width=console_width)
            measurement = Measurement.get(measure_console, measure_console.options, pretty)
            # Use the maximum width that fits the content, with some buffer for panel border rendering
            width = measurement.maximum + 4
            # print(f"width: {width}")
        if console_width is not None:
            width = min(width, console_width)
        return cls.wrap_in_panel(pretty=pretty, title=title, subtitle=subtitle, border_style=border_style, width=width)

    @classmethod
    def wrap_in_panel(
        cls,
        pretty: PrettyPrintable,
        title: TextType | None = None,
        subtitle: TextType | None = None,
        border_style: StyleType | None = None,
        width: int | None = None,
    ) -> Panel:
        return Panel(
            pretty,
            title=title,
            subtitle=subtitle,
            expand=False,
            title_align="left",
            subtitle_align="right",
            padding=(1, 1),
            border_style=border_style or "",
            highlight=True,
            width=width,
        )

    @classmethod
    def pretty_html(
        cls,
        pretty: PrettyPrintable,
        width: int = PRETTY_WIDTH_FOR_EXPORT,
    ) -> str:
        buf = StringIO()
        console = Console(record=True, file=buf, width=width, force_terminal=False)
        console.print(pretty)
        return console.export_html(inline_styles=False, clear=False, theme=EXPORT_THEME)

    @classmethod
    def pretty_svg(cls, pretty: PrettyPrintable, width: int = PRETTY_WIDTH_FOR_EXPORT) -> str:
        buf = StringIO()
        console = Console(record=True, file=buf, width=width, force_terminal=False)
        console.print(pretty)
        return console.export_svg()

    @classmethod
    def make_pretty(cls, value: Any, inner_title: str | None = None, depth: int = 0) -> PrettyPrintable:
        pretty: PrettyPrintable
        # Format the value
        if isinstance(value, PrettyPrintable):
            pretty = value
        elif isinstance(value, dict):
            # For dicts, use JSON rendering
            try:
                pretty = JSON.from_data(value, indent=4)
            except TypeError:
                json_string = kajson.dumps(value, indent=4)
                pretty = Syntax(json_string, "json", theme="monokai")
        elif isinstance(value, list):
            # For lists, build a table without headers
            list_table = Table(
                show_header=False,
                show_edge=False,
                show_lines=True,
                border_style="dim",
                padding=(0, 1),
            )
            list_table.add_column("No.", style="yellow", justify="center", width=4)
            list_table.add_column("Item", style="white")

            for idx, item in enumerate(value, start=1):  # type: ignore[arg-type]
                pretty_item = cls.make_pretty(item, inner_title=None, depth=depth + 1)
                list_table.add_row(str(idx), pretty_item)

            pretty = list_table
        elif isinstance(value, str):
            # Handle URLs specially, otherwise use Text for simple strings
            if value.startswith(("http://", "https://")):
                pretty = Text(value, style="link " + value, no_wrap=True)
            else:
                # Use Text instead of Markdown to allow proper auto-sizing
                pretty = Text(value)
        elif isinstance(value, (int, float, bool)):
            # For primitive types, convert to string
            pretty = Text(str(value))
        elif isinstance(value, PrettyRenderable):
            pretty = value.rendered_pretty(depth=depth)
        else:
            # For other types, use Pretty
            pretty = Pretty(value)

        if inner_title:
            inner_title_text = Text(str(inner_title), style="dim")
            pretty = Group(inner_title_text, pretty)

        return pretty

    @classmethod
    def pretty_print_without_rich(
        cls,
        content: str | Any,
        title: TextType | None = None,
        subtitle: TextType | None = None,
        inner_title: str | None = None,
        width: int | None = None,
        console_width: int | None = None,
    ):
        if isinstance(content, str) and content.startswith(("http://", "https://")):
            cls.pretty_print_url_without_rich(content=content, title=title, subtitle=subtitle)
            return
        title_str = str(title) if title else ""
        if subtitle:
            title_str += f"\n{subtitle!s}"
        if inner_title:
            title_str += f"\n{inner_title}"
        terminal_width = console_width or shutil.get_terminal_size().columns
        content_str = f"{content}"

        # Split title into lines if it contains newlines
        title_lines = title_str.splitlines() if title_str else []

        # Calculate max content width based on longest title line
        max_title_len = max(len(line) for line in title_lines) if title_lines else 0
        max_content_width = terminal_width - max_title_len - 8  # Accounting for frame and padding
        if width:
            max_content_width = min(max_content_width, width)
        wrapped_lines: list[str] = []
        for line in content_str.splitlines():
            while len(line) > max_content_width:
                wrapped_lines.append(line[:max_content_width])
                line = line[max_content_width:]
            wrapped_lines.append(line)

        if not wrapped_lines:
            wrapped_lines.append("")

        # Calculate frame width based on longest title line and content lines
        max_title_width = max((len(line) for line in title_lines), default=0)
        max_content_line_width = max(len(line) for line in wrapped_lines)
        frame_width = max(max_title_width + 6, max_content_line_width + 6)
        top_border = "╭" + "─" * (frame_width - 2) + "╮"
        bottom_border = "╰" + "─" * (frame_width - 2) + "╯"

        print_to_stderr(f"{BORDER_COLOR}{top_border}{RESET_FONT}")
        # Print each title line separately
        for title_line in title_lines:
            padding = " " * (frame_width - len(title_line) - 4)
            print_to_stderr(f"{BORDER_COLOR}│ {BOLD_FONT}{TITLE_COLOR}{title_line}{RESET_FONT}:{padding}{BORDER_COLOR}│{RESET_FONT}")
        for line in wrapped_lines:
            padding = " " * (frame_width - len(line) - 3)
            print_to_stderr(f"{BORDER_COLOR}│ {TEXT_COLOR}{line}{RESET_FONT}{padding}{BORDER_COLOR}│{RESET_FONT}")
        print_to_stderr(f"{BORDER_COLOR}{bottom_border}{RESET_FONT}")

    @classmethod
    def pretty_print_url_without_rich(
        cls,
        content: str | Any,
        title: TextType | None = None,
        subtitle: TextType | None = None,
    ):
        title = title or ""
        if subtitle:
            title += f" ({subtitle})"
        terminal_width = shutil.get_terminal_size().columns
        frame_width = terminal_width - 2
        top_border = "╭" + "─" * (frame_width - 2) + "╮"
        bottom_border = "╰" + "─" * (frame_width - 2) + "╯"

        print_to_stderr(f"{BORDER_COLOR}{top_border}{RESET_FONT}")
        if title:
            title_padding = " " * (frame_width - len(title) - 4)
            print_to_stderr(f"{BORDER_COLOR}│ {BOLD_FONT}{TITLE_COLOR}{title}{RESET_FONT}:{title_padding}{BORDER_COLOR}│{RESET_FONT}")
        print_to_stderr(f"{TEXT_COLOR}{content}{RESET_FONT}")
        print_to_stderr(f"{BORDER_COLOR}{bottom_border}{RESET_FONT}")



import json
from typing import Any

from json2html import json2html
from pydantic import field_validator
from rich.json import JSON
from typing_extensions import override

from pipelex.core.stuffs.stuff_content import StuffContent
from pipelex.tools.misc.markdown_utils import convert_to_markdown
from pipelex.tools.misc.pretty import PrettyPrintable


# TODO: use pipelex.tools.misc.json_utils.JsonContent to support lists in addition to dicts
class JSONContent(StuffContent):
    json_obj: dict[str, Any]

    @field_validator("json_obj", mode="before")
    @classmethod
    def check_valid_json(cls, value: dict[str, Any]) -> dict[str, Any]:
        try:
            json.dumps(value)
        except TypeError as exc:
            msg = f"json_obj is not valid JSON: {exc}"
            raise TypeError(msg) from exc
        except json.JSONDecodeError as exc:
            msg = f"json_obj is not valid JSON: {exc}"
            raise ValueError(msg) from exc
        return value

    @override
    def rendered_html(self) -> str:
        html: str = json2html.convert(  # pyright: ignore[reportAssignmentType, reportUnknownVariableType]
            json=self.json_obj,  # pyright: ignore[reportArgumentType]
            clubbing=True,
            table_attributes="",
        )
        return html

    @override
    def rendered_markdown(self, level: int = 1, is_pretty: bool = False) -> str:
        return convert_to_markdown(data=self.json_obj, level=level, is_pretty=is_pretty)

    @override
    def rendered_plain(self) -> str:
        return json.dumps(self.json_obj, indent=4)

    @override
    def rendered_json(self) -> str:
        return json.dumps(self.json_obj, indent=4)

    @override
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        return JSON.from_data(self.json_obj, indent=4)



import asyncio
from typing import TYPE_CHECKING, Any, Literal

from pydantic import field_validator, model_validator
from typing_extensions import override

from pipelex import log
from pipelex.core.concepts.concept import Concept
from pipelex.core.memory.working_memory import WorkingMemory
from pipelex.core.pipes.exceptions import PipeValidationError
from pipelex.core.pipes.inputs.exceptions import PipeInputNotFoundError
from pipelex.core.pipes.inputs.input_requirements import InputRequirements
from pipelex.core.pipes.inputs.input_requirements_factory import InputRequirementsFactory
from pipelex.core.pipes.pipe_output import PipeOutput
from pipelex.core.stuffs.stuff_factory import StuffFactory
from pipelex.hub import get_pipeline_tracker, get_required_pipe
from pipelex.libraries.pipe.exceptions import PipeNotFoundError
from pipelex.pipe_controllers.pipe_controller import PipeController
from pipelex.pipe_controllers.sub_pipe import SubPipe
from pipelex.pipe_run.exceptions import PipeRunError
from pipelex.pipe_run.pipe_run_mode import PipeRunMode
from pipelex.pipe_run.pipe_run_params import PipeRunParams
from pipelex.pipeline.job_metadata import JobMetadata
from pipelex.types import Self

if TYPE_CHECKING:
    from collections.abc import Coroutine

    from pipelex.core.stuffs.stuff import Stuff
    from pipelex.core.stuffs.stuff_content import StuffContent


class PipeParallel(PipeController):
    type: Literal["PipeParallel"] = "PipeParallel"

    parallel_sub_pipes: list[SubPipe]
    add_each_output: bool
    combined_output: Concept | None

    @field_validator("parallel_sub_pipes", mode="before")
    @classmethod
    def validate_parallel_sub_pipes(cls, parallel_sub_pipes: list[SubPipe]) -> list[SubPipe]:
        seen_output_names: set[str] = set()
        for sub_pipe in parallel_sub_pipes:
            if not sub_pipe.output_name:
                msg = f"PipeParallel '{cls.code}' sub-pipe '{sub_pipe.pipe_code}' output name not specified"
                raise ValueError(msg)
            if sub_pipe.output_name in seen_output_names:
                msg = (
                    f"PipeParallel '{cls.code}' sub-pipe '{sub_pipe.pipe_code}' output name '{sub_pipe.output_name}' "
                    "is already used by another sub-pipe"
                )
                raise ValueError(msg)
            seen_output_names.add(sub_pipe.output_name)
        return parallel_sub_pipes

    @override
    def required_variables(self) -> set[str]:
        return set()

    @override
    def needed_inputs(self, visited_pipes: set[str] | None = None) -> InputRequirements:
        if visited_pipes is None:
            visited_pipes = set()

        # If we've already visited this pipe, stop recursion
        if self.code in visited_pipes:
            return InputRequirementsFactory.make_empty()

        # Add this pipe to visited set for recursive calls
        visited_pipes_with_current = visited_pipes | {self.code}

        needed_inputs = InputRequirementsFactory.make_empty()

        for sub_pipe in self.parallel_sub_pipes:
            pipe = get_required_pipe(pipe_code=sub_pipe.pipe_code)
            # Use the centralized recursion detection
            pipe_needed_inputs = pipe.needed_inputs(visited_pipes_with_current)
            if sub_pipe.batch_params:
                try:
                    requirement = pipe_needed_inputs.get_required_input_requirement(variable_name=sub_pipe.batch_params.input_item_stuff_name)
                except PipeInputNotFoundError as exc:
                    msg = (
                        f"Batch input item named '{sub_pipe.batch_params.input_item_stuff_name}' is not "
                        f"in this Parallel Pipe '{self.code}' input requirements: {pipe_needed_inputs}"
                    )
                    raise PipeValidationError(message=msg) from exc
                needed_inputs.add_requirement(
                    variable_name=sub_pipe.batch_params.input_list_stuff_name,
                    concept=requirement.concept,
                    multiplicity=True,
                )
                for input_name, requirement in pipe_needed_inputs.items:
                    if input_name != sub_pipe.batch_params.input_item_stuff_name:
                        needed_inputs.add_requirement(input_name, requirement.concept, requirement.multiplicity)
            else:
                for input_name, requirement in pipe_needed_inputs.items:
                    needed_inputs.add_requirement(input_name, requirement.concept, requirement.multiplicity)
        return needed_inputs

    @model_validator(mode="after")
    def validate_fields_add_each_output_and_combined_output(self) -> Self:
        # Validate that either add_each_output or combined_output is set
        if not self.add_each_output and not self.combined_output:
            msg = f"PipeParallel'{self.code}'requires either add_each_output or combined_output to be set"
            raise ValueError(msg)

        return self

    @override
    def validate_inputs_static(self):
        pass

    @override
    def validate_inputs_with_library(self):
        pass

    @override
    def validate_output_static(self):
        pass

    @override
    def validate_output_with_library(self):
        pass

    @override
    def pipe_dependencies(self) -> set[str]:
        return {sub_pipe.pipe_code for sub_pipe in self.parallel_sub_pipes}

    @override
    async def _run_controller_pipe(
        self,
        job_metadata: JobMetadata,
        working_memory: WorkingMemory,
        pipe_run_params: PipeRunParams,
        output_name: str | None = None,
    ) -> PipeOutput:
        if pipe_run_params.final_stuff_code:
            log.verbose(f"PipeBatch.run_pipe() final_stuff_code: {pipe_run_params.final_stuff_code}")
            pipe_run_params.final_stuff_code = None

        tasks: list[Coroutine[Any, Any, PipeOutput]] = []

        for sub_pipe in self.parallel_sub_pipes:
            tasks.append(
                sub_pipe.run_pipe(
                    calling_pipe_code=self.code,
                    job_metadata=job_metadata,
                    working_memory=working_memory.make_deep_copy(),
                    sub_pipe_run_params=pipe_run_params.make_deep_copy(),
                ),
            )

        pipe_outputs = await asyncio.gather(*tasks)

        output_stuff_content_items: list[StuffContent] = []
        output_stuffs: dict[str, Stuff] = {}
        output_stuff_contents: dict[str, StuffContent] = {}

        # TODO: refactor this to use a specific function for this that can also be used in dry run
        for output_index, pipe_output in enumerate(pipe_outputs):
            output_stuff = pipe_output.main_stuff
            sub_pipe_output_name = self.parallel_sub_pipes[output_index].output_name
            if not sub_pipe_output_name:
                msg = "PipeParallel requires a result specified for each parallel sub pipe"
                raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)
            if self.add_each_output:
                working_memory.add_new_stuff(name=sub_pipe_output_name, stuff=output_stuff)
            output_stuff_content_items.append(output_stuff.content)
            if sub_pipe_output_name in output_stuffs:
                # TODO: check that at the blueprint / factory level
                msg = f"PipeParallel requires unique output names for each parallel sub pipe, but {sub_pipe_output_name} is already used"
                raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)
            output_stuffs[sub_pipe_output_name] = output_stuff
            if sub_pipe_output_name in output_stuff_contents:
                # TODO: check that at the blueprint / factory level
                msg = f"PipeParallel requires unique output names for each parallel sub pipe, but {sub_pipe_output_name} is already used"
                raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)
            output_stuff_contents[sub_pipe_output_name] = output_stuff.content
            log.verbose(f"PipeParallel '{self.code}': output_stuff_contents[{sub_pipe_output_name}]: {output_stuff_contents[sub_pipe_output_name]}")

        if self.combined_output:
            combined_output_stuff = StuffFactory.combine_stuffs(
                concept=self.combined_output,
                stuff_contents=output_stuff_contents,
                name=output_name,
            )
            working_memory.set_new_main_stuff(
                stuff=combined_output_stuff,
                name=output_name,
            )
            for stuff in output_stuffs.values():
                get_pipeline_tracker().add_aggregate_step(
                    from_stuff=stuff,
                    to_stuff=combined_output_stuff,
                    pipe_layer=pipe_run_params.pipe_layers,
                    comment="PipeParallel on output_stuffs",
                )
        return PipeOutput(
            working_memory=working_memory,
            pipeline_run_id=job_metadata.pipeline_run_id,
        )

    @override
    async def _dry_run_controller_pipe(
        self,
        job_metadata: JobMetadata,
        working_memory: WorkingMemory,
        pipe_run_params: PipeRunParams,
        output_name: str | None = None,
    ) -> PipeOutput:
        """Dry run implementation for PipeParallel.
        Validates that all required inputs are present and that all parallel sub-pipes can be dry run.
        """
        log.verbose(f"PipeParallel: dry run controller pipe: {self.code}")
        if pipe_run_params.run_mode != PipeRunMode.DRY:
            msg = f"PipeSequence._dry_run_controller_pipe() called with run_mode = {pipe_run_params.run_mode} in pipe {self.code}"
            raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)

        # 1. Validate that all required inputs are present in the working memory
        needed_inputs = self.needed_inputs()
        missing_input_names: list[str] = []
        for named_input_requirement in needed_inputs.named_input_requirements:
            if not working_memory.get_optional_stuff(named_input_requirement.variable_name):
                missing_input_names.append(named_input_requirement.variable_name)

        if missing_input_names:
            msg = f"Dry run failed for pipe '{self.code}' (PipeParallel): missing required inputs: {missing_input_names}"
            raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)

        # 2. Validate that all sub-pipes exist
        for sub_pipe in self.parallel_sub_pipes:
            try:
                get_required_pipe(pipe_code=sub_pipe.pipe_code)
            except PipeNotFoundError as exc:
                msg = f"Dry run failed for pipe '{self.code}' (PipeParallel): sub-pipe '{sub_pipe.pipe_code}' not found"
                raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode) from exc

        # 3. Run all sub-pipes in dry mode
        tasks: list[Coroutine[Any, Any, PipeOutput]] = []

        for sub_pipe in self.parallel_sub_pipes:
            tasks.append(
                sub_pipe.run_pipe(
                    calling_pipe_code=self.code,
                    job_metadata=job_metadata,
                    working_memory=working_memory.make_deep_copy(),
                    sub_pipe_run_params=pipe_run_params.make_deep_copy(),
                ),
            )

        pipe_outputs = await asyncio.gather(*tasks)

        # 4. Process outputs as in the regular run
        output_stuffs: dict[str, Stuff] = {}
        output_stuff_contents: dict[str, StuffContent] = {}

        for output_index, pipe_output in enumerate(pipe_outputs):
            output_stuff = pipe_output.main_stuff
            sub_pipe_output_name = self.parallel_sub_pipes[output_index].output_name
            if not sub_pipe_output_name:
                sub_pipe_code = self.parallel_sub_pipes[output_index].pipe_code
                msg = f"Dry run failed for pipe '{self.code}' (PipeParallel): sub-pipe '{sub_pipe_code}' output name not specified"
                raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)

            if self.add_each_output:
                working_memory.add_new_stuff(name=sub_pipe_output_name, stuff=output_stuff)

            if sub_pipe_output_name in output_stuffs:
                sub_pipe_code = self.parallel_sub_pipes[output_index].pipe_code
                msg = (
                    f"Dry run failed for pipe '{self.code}' (PipeParallel): sub-pipe '{sub_pipe_code}' duplicate output name '{sub_pipe_output_name}'"
                )
                raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)

            output_stuffs[sub_pipe_output_name] = output_stuff
            output_stuff_contents[sub_pipe_output_name] = output_stuff.content

        # 5. Handle combined output if specified
        if self.combined_output:
            combined_output_stuff = StuffFactory.combine_stuffs(
                concept=self.combined_output,
                stuff_contents=output_stuff_contents,
                name=output_name,
            )
            working_memory.set_new_main_stuff(
                stuff=combined_output_stuff,
                name=output_name,
            )
        return PipeOutput(
            working_memory=working_memory,
            pipeline_run_id=job_metadata.pipeline_run_id,
        )



from typing import Literal

from typing_extensions import override

from pipelex.cogt.extract.extract_setting import ExtractModelChoice
from pipelex.core.pipes.pipe_blueprint import PipeBlueprint


class PipeExtractBlueprint(PipeBlueprint):
    type: Literal["PipeExtract"] = "PipeExtract"
    pipe_category: Literal["PipeOperator"] = "PipeOperator"
    model: ExtractModelChoice | None = None
    page_images: bool | None = None
    page_image_captions: bool | None = None
    page_views: bool | None = None
    page_views_dpi: int | None = None

    @override
    def validate_inputs(self):
        nb_inputs = self.nb_inputs
        msg = (
            "Exactly one input must be provided for the PipeExtract, and it must be a pdf or an image or a concept that refines one of them."
            f"{nb_inputs} inputs were provided."
        )
        if self.inputs is None:
            raise ValueError(msg)
        if nb_inputs > 1:
            msg = f"Too many inputs provided for PipeExtract: {self.input_names}. Only one input is allowed."
            raise ValueError(msg)
        if nb_inputs < 1:
            msg = f"Missing input provided for PipeExtract: {self.input_names}. Only one input is allowed."
            raise ValueError(msg)

    @override
    def validate_output(self):
        if self.output != "Page[]":
            msg = f"PipeExtract output should be a 'native.Page[]' concept, but is {self.output}"
            raise ValueError(msg)



from pydantic import ValidationError

from pipelex.base_exceptions import PipelexError
from pipelex.kit.index_models import KitIndex
from pipelex.kit.paths import get_kit_root
from pipelex.tools.misc.toml_utils import load_toml_from_path
from pipelex.tools.typing.pydantic_utils import format_pydantic_validation_error


class KitIndexLoadingError(PipelexError):
    pass


def load_index() -> KitIndex:
    """Load and validate the kit index.toml configuration.

    Returns:
        Validated KitIndex model

    Raises:
        TomlError: If TOML parsing fails
        KitIndexLoadingError: If validation fails
    """
    index_path = get_kit_root() / "index.toml"
    data = load_toml_from_path(str(index_path))
    try:
        return KitIndex.model_validate(data)
    except ValidationError as exc:
        msg = f"Validation error in kit index at '{index_path}': {format_pydantic_validation_error(exc)}"
        raise KitIndexLoadingError(message=msg) from exc



import re

from pipelex.system.environment import EnvVarNotFoundError, get_optional_env, get_required_env
from pipelex.system.exceptions import ToolError
from pipelex.tools.secrets.secrets_errors import SecretNotFoundError
from pipelex.tools.secrets.secrets_provider_abstract import SecretsProviderAbstract
from pipelex.types import StrEnum


class VarNotFoundError(ToolError):
    def __init__(self, var_name: str, message: str):
        self.var_name = var_name
        super().__init__(message)


class VarFallbackPatternError(ToolError):
    pass


class UnknownVarPrefixError(ToolError):
    """Raised when an unknown variable prefix is used in variable substitution."""

    def __init__(self, var_name: str, message: str):
        self.var_name = var_name
        super().__init__(message)


class VarPrefix(StrEnum):
    """Variable prefix types for variable substitution."""

    ENV = "env"
    SECRET = "secret"


def substitute_vars(content: str, secrets_provider: SecretsProviderAbstract) -> str:
    """Substitute variable placeholders with values from environment variables or secrets.

    Supports the following placeholder formats:
    - ${VAR_NAME} -> use secrets provider by default
    - ${env:ENV_VAR_NAME} -> force use environment variable
    - ${secret:SECRET_NAME} -> force use secrets provider
    - ${env:ENV_VAR_NAME|secret:SECRET_NAME} -> try env first, then secret as fallback

    Args:
        content: Text content with variable placeholders
        secrets_provider: The secrets provider to use for secret lookups

    Returns:
        Content with variables substituted

    Raises:
        VarNotFoundError: If required variable is missing from all specified sources

    """

    def replace_var(match: re.Match[str]) -> str:
        var_spec = match.group(1)

        # Check if it's a fallback pattern (contains |)
        if "|" in var_spec:
            return _handle_fallback_pattern(var_spec, secrets_provider)

        # Check if it has a prefix (env: or secret:)
        if ":" in var_spec:
            prefix_str, var_name = var_spec.split(":", 1)
            prefix_str = prefix_str.strip()

            try:
                prefix = VarPrefix(prefix_str)
            except ValueError as exc:
                msg = f"Unknown variable prefix: '{prefix_str}'"
                raise UnknownVarPrefixError(
                    var_name=var_name,
                    message=msg,
                ) from exc

            match prefix:
                case VarPrefix.ENV:
                    return _get_env_var(var_name)
                case VarPrefix.SECRET:
                    return _get_secret(var_name, secrets_provider)
        else:
            # Default behavior: use secrets provider
            return _get_secret(var_spec, secrets_provider)

    # Pattern matches ${VAR_NAME} or ${prefix:VAR_NAME} or ${env:VAR|secret:VAR}
    # Restrict to not match across newlines, quotes, or nested braces
    pattern = r"\$\{([^}\n\"'$]+)\}"
    return re.sub(pattern, replace_var, content)


def _handle_fallback_pattern(var_spec: str, secrets_provider: SecretsProviderAbstract) -> str:
    """Handle fallback pattern like 'env:VAR|secret:VAR'."""
    parts = [part.strip() for part in var_spec.split("|")]

    for part in parts:
        if ":" in part:
            prefix_str, var_name = part.split(":", 1)
            prefix_str = prefix_str.strip()

            try:
                prefix = VarPrefix(prefix_str)
            except ValueError as exc:
                msg = f"Unknown variable prefix: '{prefix_str}'"
                raise UnknownVarPrefixError(
                    var_name=var_name,
                    message=msg,
                ) from exc

            match prefix:
                case VarPrefix.ENV:
                    value = get_optional_env(var_name)
                    if value is not None:
                        return value
                case VarPrefix.SECRET:
                    try:
                        return secrets_provider.get_secret(secret_id=var_name)
                    except SecretNotFoundError:
                        continue  # Try next option
        else:
            # No prefix, try as secret
            try:
                return secrets_provider.get_secret(secret_id=part)
            except SecretNotFoundError:
                continue  # Try next option
    msg = f"Could not get variable from fallback pattern: {var_spec}"
    raise VarFallbackPatternError(message=msg)


def _get_env_var(var_name: str) -> str:
    """Get environment variable, raising VarNotFoundError if not found."""
    try:
        return get_required_env(var_name)
    except EnvVarNotFoundError as exc:
        msg = f"Could not get variable '{var_name}': {exc!s}"
        raise VarNotFoundError(message=msg, var_name=var_name) from exc


def _get_secret(secret_name: str, secrets_provider: SecretsProviderAbstract) -> str:
    """Get secret, raising VarNotFoundError if not found."""
    try:
        return secrets_provider.get_secret(secret_id=secret_name)
    except SecretNotFoundError as exc:
        msg = f"Could not get variable '{secret_name}': {exc!s}"
        raise VarNotFoundError(message=msg, var_name=secret_name) from exc



from typing import Any, Generic

from json2html import json2html
from rich.pretty import Pretty
from rich.table import Table
from typing_extensions import override

from pipelex.cogt.templating.templating_style import TextFormat
from pipelex.core.stuffs.stuff_content import StuffContent, StuffContentType
from pipelex.tools.misc.pretty import MAX_RENDER_DEPTH, PrettyPrintable, PrettyPrinter


class ListContent(StuffContent, Generic[StuffContentType]):
    items: list[StuffContentType]

    @property
    def nb_items(self) -> int:
        return len(self.items)

    def get_items(self, item_type: type[StuffContent]) -> list[StuffContent]:
        return [item for item in self.items if isinstance(item, item_type)]

    @property
    @override
    def short_desc(self) -> str:
        nb_items = len(self.items)
        if nb_items == 0:
            return "empty list"
        if nb_items == 1:
            return f"list of 1 {self.items[0].__class__.__name__}"
        item_classes: list[str] = [item.__class__.__name__ for item in self.items]
        item_classes_set = set(item_classes)
        nb_classes = len(item_classes_set)
        if nb_classes == 1:
            return f"list of {len(self.items)} {item_classes[0]}s"
        elif nb_items == nb_classes:
            return f"list of {len(self.items)} items of different types"
        else:
            return f"list of {len(self.items)} items of {nb_classes} different types"

    @property
    def _single_class_name(self) -> str | None:
        item_classes: list[str] = [item.__class__.__name__ for item in self.items]
        item_classes_set = set(item_classes)
        nb_classes = len(item_classes_set)
        if nb_classes == 1:
            return item_classes[0]
        else:
            return None

    @override
    def model_dump(self, *args: Any, **kwargs: Any):
        obj_dict = super().model_dump(*args, **kwargs)
        obj_dict["items"] = [item.model_dump(*args, **kwargs) for item in self.items]
        return obj_dict

    @override
    def rendered_plain(self) -> str:
        return self.rendered_markdown()

    @override
    def rendered_html(self) -> str:
        list_dump = [item.smart_dump() for item in self.items]

        html: str = json2html.convert(  # pyright: ignore[reportAssignmentType, reportUnknownVariableType]
            json=list_dump,  # pyright: ignore[reportArgumentType]
            clubbing=True,
            table_attributes="",
        )
        return html

    @override
    def rendered_markdown(self, level: int = 1, is_pretty: bool = False) -> str:
        rendered = ""
        if self._single_class_name == "TextContent":
            for item in self.items:
                rendered += f" • {item}\n"
        else:
            for item_index, item in enumerate(self.items):
                rendered += f"\n • item #{item_index + 1}:\n\n"
                rendered += item.rendered_str(text_format=TextFormat.MARKDOWN)
                rendered += "\n"
        return rendered

    @override
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        # Check if we've exceeded maximum depth - fall back to Pretty rendering
        # Pretty shows the Python object structure beautifully, just like when calling pretty_print(stuff)
        if depth >= MAX_RENDER_DEPTH:
            return Pretty(self)

        table = Table(
            title=title,
            show_header=False,
            show_edge=False,
            show_lines=True,
            border_style="white",
            width=PrettyPrinter.pretty_width(depth=depth),
        )
        table.add_column("No.", style="yellow", justify="center", width=6)
        table.add_column("Content", style="white")

        for item_index, item in enumerate(self.items):
            item_number = str(item_index + 1)
            item_content = item.rendered_pretty(depth=depth + 1)
            table.add_row(item_number, item_content)

        return table



from typing import Literal

from pydantic import field_validator
from typing_extensions import override

from pipelex.core.memory.working_memory import WorkingMemory
from pipelex.core.pipes.exceptions import PipeValidationError, PipeValidationErrorType
from pipelex.core.pipes.inputs.exceptions import PipeInputNotFoundError
from pipelex.core.pipes.inputs.input_requirements import InputRequirements
from pipelex.core.pipes.inputs.input_requirements_factory import InputRequirementsFactory
from pipelex.core.pipes.pipe_output import PipeOutput
from pipelex.hub import get_concept_library, get_required_pipe
from pipelex.pipe_controllers.exceptions import PipeControllerOutputConceptMismatchError
from pipelex.pipe_controllers.parallel.pipe_parallel import PipeParallel
from pipelex.pipe_controllers.pipe_controller import PipeController
from pipelex.pipe_controllers.sequence.exceptions import PipeSequenceValueError
from pipelex.pipe_controllers.sub_pipe import SubPipe
from pipelex.pipe_run.exceptions import PipeRunError, PipeRunParamsError
from pipelex.pipe_run.pipe_run_params import PipeRunParams
from pipelex.pipeline.job_metadata import JobMetadata


class PipeSequence(PipeController):
    type: Literal["PipeSequence"] = "PipeSequence"
    sequential_sub_pipes: list[SubPipe]

    @override
    def required_variables(self) -> set[str]:
        return set()

    @field_validator("sequential_sub_pipes", mode="after")
    @classmethod
    def validate_sequential_sub_pipes(cls, value: list[SubPipe]) -> list[SubPipe]:
        if not value:
            msg = f"PipeSequence '{cls.code}' requires at least one sub-pipe"
            raise ValueError(msg)
        return value

    @override
    def validate_inputs_static(self):
        pass

    @override
    def validate_inputs_with_library(self):
        pass

    @override
    def validate_output_static(self):
        pass

    @override
    def validate_output_with_library(self):
        """Validate the output for the pipe sequence.
        The output of the pipe sequence should match the output of the last step.
        """
        last_step_pipe_code = self.sequential_sub_pipes[-1].pipe_code
        last_step_pipe = get_required_pipe(pipe_code=last_step_pipe_code)
        last_step_output_concept = last_step_pipe.output
        if not get_concept_library().is_compatible(tested_concept=last_step_output_concept, wanted_concept=self.output):
            msg = (
                f"PipeSequence concept mismatch: the output concept '{last_step_output_concept.concept_string}' "
                f"of the last step '{last_step_pipe_code}' of sequence pipe '{self.code}' "
                f"is not compatible with the output concept '{self.output.concept_string}' of the sequence."
            )
            raise PipeValidationError(
                message=msg,
                error_type=PipeValidationErrorType.INADEQUATE_OUTPUT_CONCEPT,
                domain=self.domain,
                pipe_code=self.code,
                provided_concept_code=last_step_output_concept.concept_string,
                required_concept_codes=[self.output.concept_string],
            )

    @override
    def needed_inputs(self, visited_pipes: set[str] | None = None) -> InputRequirements:
        if visited_pipes is None:
            visited_pipes = set()

        # If we've already visited this pipe, stop recursion
        if self.code in visited_pipes:
            return InputRequirementsFactory.make_empty()

        # Add this pipe to visited set for recursive calls
        visited_pipes_with_current = visited_pipes | {self.code}

        needed_inputs = InputRequirementsFactory.make_empty()
        generated_outputs: set[str] = set()

        for sequential_sub_pipe in self.sequential_sub_pipes:
            sub_pipe = get_required_pipe(pipe_code=sequential_sub_pipe.pipe_code)
            # Use the centralized recursion detection
            sub_pipe_needed_inputs = sub_pipe.needed_inputs(visited_pipes_with_current)

            if isinstance(sub_pipe, PipeParallel) and sub_pipe.add_each_output:
                for sub_parallel_pipe in sub_pipe.parallel_sub_pipes:
                    if (sub_pipe.add_each_output and sub_parallel_pipe.output_name) or sub_parallel_pipe.output_name:
                        generated_outputs.add(sub_parallel_pipe.output_name)

            if sequential_sub_pipe.batch_params:
                if sequential_sub_pipe.batch_params.input_list_stuff_name not in generated_outputs:
                    try:
                        requirement = sub_pipe_needed_inputs.get_required_input_requirement(
                            variable_name=sequential_sub_pipe.batch_params.input_item_stuff_name
                        )
                    except PipeInputNotFoundError as exc:
                        msg = (
                            f"Batch input item named '{sequential_sub_pipe.batch_params.input_item_stuff_name}' is not "
                            f"in this PipeSequence '{self.code}' input requirements: {sub_pipe_needed_inputs}"
                        )
                        raise PipeSequenceValueError(msg) from exc
                    needed_inputs.add_requirement(
                        variable_name=sequential_sub_pipe.batch_params.input_list_stuff_name,
                        concept=requirement.concept,
                        multiplicity=True,
                    )
                    for input_name, requirement in sub_pipe_needed_inputs.items:
                        if input_name != sequential_sub_pipe.batch_params.input_item_stuff_name and input_name not in generated_outputs:
                            needed_inputs.add_requirement(input_name, requirement.concept, requirement.multiplicity)
            else:
                for input_name, requirement in sub_pipe_needed_inputs.items:
                    if input_name not in generated_outputs:
                        needed_inputs.add_requirement(input_name, requirement.concept, requirement.multiplicity)

            # Add this step's output to generated outputs
            if sequential_sub_pipe.output_name:
                generated_outputs.add(sequential_sub_pipe.output_name)
        return needed_inputs

    @override
    def pipe_dependencies(self) -> set[str]:
        return {sub_pipe.pipe_code for sub_pipe in self.sequential_sub_pipes}

    @override
    async def _run_controller_pipe(
        self,
        job_metadata: JobMetadata,
        working_memory: WorkingMemory,
        pipe_run_params: PipeRunParams,
        output_name: str | None = None,
    ) -> PipeOutput:
        pipe_run_params.push_pipe_layer(pipe_code=self.code)
        if pipe_run_params.is_multiple_output_required:
            msg = f"{self.__class__.__name__} does not support multiple outputs, got output_multiplicity = {pipe_run_params.output_multiplicity}"
            raise PipeRunParamsError(msg)

        evolving_memory = working_memory

        for sub_pipe_index, sub_pipe in enumerate(self.sequential_sub_pipes):
            # Only the last step should apply the final_stuff_code
            if sub_pipe_index == len(self.sequential_sub_pipes) - 1:
                sub_pipe_run_params = pipe_run_params.model_copy()
            else:
                sub_pipe_run_params = pipe_run_params.model_copy(update=({"final_stuff_code": None}))
            pipe_output = await sub_pipe.run_pipe(
                calling_pipe_code=self.code,
                working_memory=evolving_memory,
                job_metadata=job_metadata,
                sub_pipe_run_params=sub_pipe_run_params,
            )
            evolving_memory = pipe_output.working_memory
        return PipeOutput(
            working_memory=evolving_memory,
            pipeline_run_id=job_metadata.pipeline_run_id,
        )

    @override
    async def _dry_run_controller_pipe(
        self,
        job_metadata: JobMetadata,
        working_memory: WorkingMemory,
        pipe_run_params: PipeRunParams,
        output_name: str | None = None,
    ) -> PipeOutput:
        if not pipe_run_params.run_mode.is_dry:
            msg = f"PipeSequence._dry_run_controller_pipe() called with run_mode = {pipe_run_params.run_mode} in pipe {self.code}"
            raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)
        # Verify the output of this pipe is matching the output of the last step.
        concept_of_last_step = get_required_pipe(pipe_code=self.sequential_sub_pipes[-1].pipe_code).output
        # if self.output.concept_string != concept_string_of_last_step:
        if not get_concept_library().is_compatible(tested_concept=concept_of_last_step, wanted_concept=self.output):
            msg = f"""PipeSequence concept mismatch:
the output concept '{concept_of_last_step.concept_string}' of the last step '{self.sequential_sub_pipes[-1].pipe_code}'
of sequence pipe '{self.code}' is not compatible with the output concept '{self.output.concept_string}' of the sequence.
"""
            raise PipeControllerOutputConceptMismatchError(
                message=msg, tested_concept=concept_of_last_step.concept_string, wanted_concept=self.output.concept_string
            )
        return await self._run_controller_pipe(
            job_metadata=job_metadata,
            working_memory=working_memory,
            pipe_run_params=pipe_run_params,
            output_name=output_name,
        )



from typing import Literal

from pydantic import Field, field_validator, model_validator
from rich.console import Group
from rich.table import Table
from rich.text import Text
from typing_extensions import override

from pipelex.builder.pipe.pipe_spec import PipeSpec
from pipelex.builder.pipe.sub_pipe_spec import SubPipeSpec
from pipelex.core.concepts.validation import validate_concept_string_or_code
from pipelex.pipe_controllers.parallel.pipe_parallel_blueprint import PipeParallelBlueprint
from pipelex.tools.misc.pretty import PrettyPrintable
from pipelex.types import Self


class PipeParallelSpec(PipeSpec):
    """Spec for parallel pipe execution in the Pipelex framework.

    PipeParallel enables concurrent execution of multiple pipes, improving performance
    for independent operations. All parallel pipes receive the same input context
    and their outputs can be combined or kept separate.

    Validation Rules:
        1. Parallels list must not be empty.
        2. Each parallel step must be a valid SubPipeSpec.
        3. combined_output, when specified, must be a valid ConceptCode in PascalCase.
        4. Pipe codes in parallels must reference existing pipes (snake_case).

    """

    type: Literal["PipeParallel"] = "PipeParallel"
    pipe_category: Literal["PipeController"] = "PipeController"
    parallels: list[SubPipeSpec] = Field(description="List of SubPipeSpec instances to execute concurrently.")
    add_each_output: bool = Field(description="Whether to include individual pipe outputs in the combined result.")
    combined_output: str | None = Field(default=None, description="Optional ConceptCode in PascalCasefor the combined output structure.")

    @field_validator("combined_output", mode="before")
    @classmethod
    def validate_combined_output(cls, combined_output: str) -> str:
        if combined_output:
            validate_concept_string_or_code(concept_string_or_code=combined_output)
        return combined_output

    @model_validator(mode="after")
    def validate_output_options(self) -> Self:
        if not self.add_each_output and not self.combined_output:
            msg = (
                "PipeParallel requires either add_each_output to be True or combined_output to be set, "
                "or both, otherwise the pipe won't output anything"
            )
            raise ValueError(msg)
        return self

    @override
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        # Get base pipe information from parent
        base_group = super().rendered_pretty(title=title, depth=depth)

        # Create a group combining base info with parallel-specific details
        parallel_group = Group()
        parallel_group.renderables.append(base_group)

        # Add parallel configuration
        parallel_group.renderables.append(Text())  # Blank line
        parallel_group.renderables.append(Text.from_markup(f"Add Each Output: [bold yellow]{self.add_each_output}[/bold yellow]"))
        if self.combined_output:
            parallel_group.renderables.append(Text.from_markup(f"Combined Output: [bold green]{self.combined_output}[/bold green]"))

        # Add parallel branches as a table
        parallel_group.renderables.append(Text())  # Blank line
        parallels_table = Table(
            title="Parallel Branches:",
            title_justify="left",
            title_style="not italic",
            show_header=True,
            header_style="dim",
            show_edge=True,
            show_lines=True,
            border_style="dim",
        )
        parallels_table.add_column("Branch", style="dim", width=6, justify="right")
        parallels_table.add_column("Pipe", style="red")
        parallels_table.add_column("Result name", style="cyan")

        for idx, parallel in enumerate(self.parallels, start=1):
            parallels_table.add_row(str(idx), parallel.pipe_code, parallel.result)

        parallel_group.renderables.append(parallels_table)

        return parallel_group

    @override
    def to_blueprint(self) -> PipeParallelBlueprint:
        base_blueprint = super().to_blueprint()
        core_parallels = [parallel.to_blueprint() for parallel in self.parallels]
        return PipeParallelBlueprint(
            description=base_blueprint.description,
            inputs=base_blueprint.inputs,
            output=base_blueprint.output,
            type=self.type,
            pipe_category=self.pipe_category,
            parallels=core_parallels,
            add_each_output=self.add_each_output,
            combined_output=self.combined_output,
        )
